"""
Derived from https://github.com/chakki-works/seqeval/blob/master/seqeval/metrics/sequence_labeling.py

Produce standard evaluation metrics for sequence labeling tasks (precision, recall, f-score at field level).

From the seqeval version, we modify:

* the classification_report() method so that micro average is computed over
actual total predictions and not weighted by number of expected pred in each
label - otherwise average micro f1 scores are not consistent with f1_score()

* an evaluation report is provided both as a string for simple console output
and as a map for further processing (e.g. n-fold average)

"""

from __future__ import absolute_import, division, print_function

from collections import defaultdict

import numpy as np


def get_entities(seq):
    """Gets entities from sequence.
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        >>> from seqeval.metrics.sequence_labeling import get_entities
        >>> seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        >>> get_entities(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ["O"]]

    prev_tag = "O"
    prev_type = ""
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ["O"]):
        tag = chunk[0]
        # type_ = chunk.split('-')[-1]
        type_ = "-".join(chunk.split("-")[1:])

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_

    return chunks


def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == "E":
        chunk_end = True
    if prev_tag == "S":
        chunk_end = True

    if prev_tag == "B" and tag == "B":
        chunk_end = True
    if prev_tag == "B" and tag == "S":
        chunk_end = True
    if prev_tag == "B" and tag == "O":
        chunk_end = True
    if prev_tag == "I" and tag == "B":
        chunk_end = True
    if prev_tag == "I" and tag == "S":
        chunk_end = True
    if prev_tag == "I" and tag == "O":
        chunk_end = True

    if prev_tag != "O" and prev_tag != "." and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == "B":
        chunk_start = True
    if tag == "S":
        chunk_start = True

    if prev_tag == "E" and tag == "E":
        chunk_start = True
    if prev_tag == "E" and tag == "I":
        chunk_start = True
    if prev_tag == "S" and tag == "E":
        chunk_start = True
    if prev_tag == "S" and tag == "I":
        chunk_start = True
    if prev_tag == "O" and tag == "E":
        chunk_start = True
    if prev_tag == "O" and tag == "I":
        chunk_start = True

    if tag != "O" and tag != "." and prev_type != type_:
        chunk_start = True

    return chunk_start


def f1_score(y_true, y_pred, average="micro"):
    """Compute the F1 score.
    The F1 score can be interpreted as a weighted average of the precision and
    recall, where an F1 score reaches its best value at 1 and worst score at 0.
    The relative contribution of precision and recall to the F1 score are
    equal. The formula for the F1 score is::
        F1 = 2 * (precision * recall) / (precision + recall)
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a tagger.
    Returns:
        score : float.
    Example:
        >>> from seqeval.metrics import f1_score
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> f1_score(y_true, y_pred)
        0.50
    """
    true_entities = set(get_entities(y_true))
    pred_entities = set(get_entities(y_pred))

    nb_correct = len(true_entities & pred_entities)
    nb_pred = len(pred_entities)
    nb_true = len(true_entities)

    p = nb_correct / nb_pred if nb_pred > 0 else 0
    r = nb_correct / nb_true if nb_true > 0 else 0
    score = 2 * p * r / (p + r) if p + r > 0 else 0

    return score


def accuracy_score(y_true, y_pred):
    """Accuracy classification score.
    In multilabel classification, this function computes subset accuracy:
    the set of labels predicted for a sample must *exactly* match the
    corresponding set of labels in y_true.
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a tagger.
    Returns:
        score : float.
    Example:
        >>> from seqeval.metrics import accuracy_score
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> accuracy_score(y_true, y_pred)
        0.80
    """
    if any(isinstance(s, list) for s in y_true):
        y_true = [item for sublist in y_true for item in sublist]
        y_pred = [item for sublist in y_pred for item in sublist]

    nb_correct = sum(y_t == y_p for y_t, y_p in zip(y_true, y_pred))
    nb_true = len(y_true)

    score = nb_correct / nb_true

    return score


def precision_score(y_true, y_pred, average="micro"):
    """Compute the precision.
    The precision is the ratio ``tp / (tp + fp)`` where ``tp`` is the number of
    true positives and ``fp`` the number of false positives. The precision is
    intuitively the ability of the classifier not to label as positive a sample.
    The best value is 1 and the worst value is 0.
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a tagger.
    Returns:
        score : float.
    Example:
        >>> from seqeval.metrics import precision_score
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> precision_score(y_true, y_pred)
        0.50
    """
    true_entities = set(get_entities(y_true))
    pred_entities = set(get_entities(y_pred))

    nb_correct = len(true_entities & pred_entities)
    nb_pred = len(pred_entities)

    score = nb_correct / nb_pred if nb_pred > 0 else 0

    return score


def recall_score(y_true, y_pred, average="micro"):
    """Compute the recall.
    The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
    true positives and ``fn`` the number of false negatives. The recall is
    intuitively the ability of the classifier to find all the positive samples.
    The best value is 1 and the worst value is 0.
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a tagger.
    Returns:
        score : float.
    Example:
        >>> from seqeval.metrics import recall_score
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> recall_score(y_true, y_pred)
        0.50
    """
    true_entities = set(get_entities(y_true))
    pred_entities = set(get_entities(y_pred))

    nb_correct = len(true_entities & pred_entities)
    nb_true = len(true_entities)

    score = nb_correct / nb_true if nb_true > 0 else 0

    return score


def classification_report(y_true, y_pred, digits=2):
    """Build a text report showing the main classification metrics.
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a classifier.
        digits : int. Number of digits for formatting output floating point values.
    Returns:
        report : string. Text summary of the precision, recall, F1 score and support number for each class.
        evaluation: map. Map with three keys:
            - 'labels', containing a map for all class with precision, recall, F1 score and support number,
            - 'micro' with precision, recall, F1 score and support number micro average
            - 'macro' with precision, recall, F1 score and support number macro average
    Examples:
        >>> from seqeval.metrics import classification_report
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> eval, _ = classification_report(y_true, y_pred)
        >>> print(eval)
                     precision    recall  f1-score   support
        <BLANKLINE>
               MISC       0.00      0.00      0.00         1
                PER       1.00      1.00      1.00         1
        <BLANKLINE>
        avg / total       0.50      0.50      0.50         2
        <BLANKLINE>
    """
    evaluation = compute_metrics(y_true, y_pred)

    return get_report(evaluation, digits=digits), evaluation


def compute_metrics(y_true, y_pred):
    """Build a text report showing the main classification metrics.
    Args:
        y_true : 2d array. Ground truth (correct) target values.
        y_pred : 2d array. Estimated targets as returned by a classifier.
    Returns:
        report : string. Text summary of the precision, recall, F1 score and support number for each class.
        evaluation: map. Map with three keys:
            - 'labels', containing a map for all class with precision, recall, F1 score and support number,
            - 'micro' with precision, recall, F1 score and support number micro average
            - 'macro' with precision, recall, F1 score and support number macro average
    Examples:
        >>> from seqeval.metrics import classification_report
        >>> y_true = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> y_pred = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        >>> eval = compute_metrics(y_true, y_pred)
        >>> print(eval)

    """
    true_entities = set(get_entities(y_true))
    pred_entities = set(get_entities(y_pred))

    d1 = defaultdict(set)
    d2 = defaultdict(set)
    for e in true_entities:
        d1[e[0]].add((e[1], e[2]))

    for e in pred_entities:
        d2[e[0]].add((e[1], e[2]))

    evaluation = {"labels": {}}

    ps, rs, f1s, s = [], [], [], []
    total_nb_correct = 0
    total_nb_pred = 0
    total_nb_true = 0
    for type_name, true_entities in d1.items():
        pred_entities = d2[type_name]
        nb_correct = len(true_entities & pred_entities)
        nb_pred = len(pred_entities)
        nb_true = len(true_entities)

        p = nb_correct / nb_pred if nb_pred > 0 else 0
        r = nb_correct / nb_true if nb_true > 0 else 0
        f1 = 2 * p * r / (p + r) if p + r > 0 else 0

        eval_block = {}
        eval_block["precision"] = p
        eval_block["recall"] = r
        eval_block["f1"] = f1
        eval_block["support"] = nb_true

        evaluation["labels"][type_name] = eval_block

        ps.append(p)
        rs.append(r)
        f1s.append(f1)
        s.append(nb_true)

        total_nb_correct += nb_correct
        total_nb_true += nb_true
        total_nb_pred += nb_pred

    # compute averages

    # marco aveage
    # macro_eval_block = {
    #     "precision": np.average(ps, weights=s),
    #     "recall": np.average(rs, weights=s),
    #     "f1": np.average(f1s, weights=s),
    #     "support": np.sum(s)
    # }
    #
    # evaluation['macro'] = macro_eval_block

    # micro average
    micro_precision = total_nb_correct / total_nb_pred if total_nb_pred > 0 else 0
    micro_recall = total_nb_correct / total_nb_true if total_nb_true > 0 else 0
    micro_f1 = (
        2 * micro_precision * micro_recall / (micro_precision + micro_recall)
        if micro_precision + micro_recall > 0
        else 0
    )

    micro_eval_block = {"precision": micro_precision, "recall": micro_recall, "f1": micro_f1, "support": np.sum(s)}
    evaluation["micro"] = micro_eval_block

    return evaluation


def get_report(evaluation, digits=2, include_avgs=["micro"]):
    """
    Calculate the report from the evaluation metrics.

    :param evaluation: the map for evaluation containing three keys:
        - 'labels', a map of maps contains the values of precision, recall, f1 and support for each label
        - 'macro' and 'micro', containing the micro and macro average, respectively (precision, recall, f1 and support)

    :param digits: the number of digits to use in the report
    :param include_avgs: the average to include in the report, default: 'micro'
    :return:
    """
    name_width = max([len(e) for e in evaluation.keys()])

    last_line_heading = {"micro": "all (micro avg.)", "macro": "all (macro avg.)"}
    width = max(name_width, len(last_line_heading["micro"]), digits)

    headers = ["precision", "recall", "f1-score", "support"]
    head_fmt = "{:>{width}s} " + " {:>9}" * len(headers)
    report = head_fmt.format("", *headers, width=width)
    report += "\n\n"

    row_fmt = "{:>{width}s} " + " {:>9.{digits}f}" * 3 + " {:>9}\n"

    block = evaluation["labels"]
    labels = sorted(block.keys())

    for label in labels:
        p = block[label]["precision"]
        r = block[label]["recall"]
        f1 = block[label]["f1"]
        s = block[label]["support"]

        report += row_fmt.format(*[label, p, r, f1, s], width=width, digits=digits)

    report += "\n"
    for average in include_avgs:
        avg = evaluation[average]
        report += row_fmt.format(
            last_line_heading[average],
            avg["precision"],
            avg["recall"],
            avg["f1"],
            avg["support"] if "support" in avg else "",
            width=width,
            digits=digits,
        )

    return report



import itertools
import json
import logging
import re
from typing import Iterable, List, Set

import numpy as np

from delft.sequenceLabelling.config import ModelConfig

LOGGER = logging.getLogger(__name__)

from sklearn.base import BaseEstimator, TransformerMixin

UNK = "<UNK>"
PAD = "<PAD>"

case_index = {
    "<PAD>": 0,
    "numeric": 1,
    "allLower": 2,
    "allUpper": 3,
    "initialUpper": 4,
    "other": 5,
    "mainly_numeric": 6,
    "contains_digit": 7,
}


def calculate_cardinality(feature_vector, indices=None):
    """
    Calculate cardinality of each features

    :param feature_vector: three dimensional vector with features
    :param indices: list of indices of the features to be extracted
    :return: a map where each key is the index of the feature and the value is a map feature_value,
    value_index.
    For example
     [(0, {'feature1': 1, 'feature2': 2})]

     indicates that the feature is at index 0 and has two values, features1 and features2 with two
     unique index.

     NOTE: the features are indexed from 1 to n + 1. The 0 value is reserved as padding
    """
    columns_length = []
    index = 0
    if not len(feature_vector) > 0:
        return []

    for index_column in range(index, len(feature_vector[0][0])):
        if indices and index_column not in indices:
            index += 1
            continue

        values = set()
        for index_document in range(0, len(feature_vector)):
            for index_row in range(0, len(feature_vector[index_document])):
                value = feature_vector[index_document][index_row][index_column]
                if value != " ":
                    values.add(value)

        values = sorted(values)
        values_cardinality = len(values)

        values_list = list(values)
        values_to_int = {}
        for val_num in range(0, values_cardinality):
            # We reserve the 0 for the unseen features so the indexes will go from 1 to cardinality + 1
            values_to_int[values_list[val_num]] = val_num + 1

        columns_length.append((index, values_to_int))
        index += 1

    return columns_length


def cardinality_to_index_map(columns_length, features_max_vector_size):
    # Filter out the columns that are not fitting
    columns_index = []
    for index, column_content_cardinality in columns_length:
        if len(column_content_cardinality) <= features_max_vector_size:
            columns_index.append((index, column_content_cardinality))
    # print(columns_index)
    features_max_vector_size * len(columns_index) + 1

    index_list = [ind[0] for ind in columns_index if ind[0] >= 0]
    val_to_int_map = {
        value[0]: {
            val_features: idx_features + (index * features_max_vector_size)
            for val_features, idx_features in value[1].items()
        }
        for index, value in enumerate(columns_index)
    }

    return index_list, val_to_int_map


def reduce_features_to_indexes(feature_vector, features_max_vector_size, indices=None):
    cardinality = calculate_cardinality(feature_vector, indices=indices)
    index_list, map_to_integers = cardinality_to_index_map(cardinality, features_max_vector_size)

    return index_list, map_to_integers


def reduce_features_vector(feature_vector, features_max_vector_size):
    """
    Reduce the features vector.
    First it calculates the cardinality for each value that each feature can assume, then
    removes features with cardinality above features_max_vector_size.
    Finally it assign indices values for each features, assuming 0 as padding (feature value out of bound or invalid),
    and values from 1 to features_max_vector_size * number of features.

    :param feature_vector: feature vector to be reduced
    :param features_max_vector_size maximum size of the one-hot-encoded values
    :return:
    """

    # Compute frequencies for each column
    columns_length = []
    index = 0
    if not len(feature_vector) > 0:
        return []

    for index_column in range(index, len(feature_vector[0])):
        values = set()
        for index_row in range(0, len(feature_vector)):
            value = feature_vector[index_row][index_column]
            if value != " ":
                values.add(value)

        values_cardinality = len(values)

        values_list = list(values)
        values_to_int = {}
        for val_num in range(0, values_cardinality):
            values_to_int[values_list[val_num]] = val_num

        columns_length.append((index, values_to_int))
        index += 1
        # print("Column: " + str(index_column) + " Len:  " + str(len(values)))

    # Filter out the columns that are not fitting
    columns_index = []
    for index, column_content_cardinality in columns_length:
        if len(column_content_cardinality) <= features_max_vector_size:
            columns_index.append((index, column_content_cardinality))
    # print(columns_index)
    index_list = [ind[0] for ind in columns_index if ind[0]]

    # Assign indexes to each feature value
    reduced_features_vector = []
    for index_row in range(0, len(feature_vector)):
        reduced_features_vector.append(
            [feature_vector[index_row][index_column] for index, index_column in enumerate(index_list)]
        )

    return reduced_features_vector


def get_map_to_index(X, features_indices, features_vector_size):
    pass


def to_dict(
    value_list_batch: List[list],
    feature_indices: Set[int] = None,
    features_vector_size: int = ModelConfig.DEFAULT_FEATURES_VOCABULARY_SIZE,
):
    if not feature_indices:
        matrix = reduce_features_vector(value_list_batch, features_vector_size)

        return [{index: value for index, value in enumerate(value_list)} for value_list in matrix]
    else:
        return [
            {index: value for index, value in enumerate(value_list) if index in feature_indices}
            for value_list in value_list_batch
        ]


class FeaturesPreprocessor(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        features_indices: Iterable[int] = None,
        features_vocabulary_size=ModelConfig.DEFAULT_FEATURES_VOCABULARY_SIZE,
        features_map_to_index=None,
    ):
        # feature_indices_set = None
        if features_map_to_index is None:
            features_map_to_index = {}
        self.features_vocabulary_size = features_vocabulary_size
        self.features_indices = features_indices

        # List of mappings to integers (corresponding to each feature column) of features values
        # This value could be provided (model has been loaded) or not (first-time-training)
        self.features_map_to_index = features_map_to_index

    def fit(self, X):
        if not self.features_indices:
            indexes, mapping = reduce_features_to_indexes(X, self.features_vocabulary_size)
        else:
            indexes, mapping = reduce_features_to_indexes(
                X, self.features_vocabulary_size, indices=self.features_indices
            )

        self.features_map_to_index = mapping
        self.features_indices = indexes
        return self

    def transform(self, X, extend=False):
        """
        Transform the features into a vector, return the vector and the extracted number of features

        :param extend: when set to true it's adding an additional empty feature list in the sequence.
        """
        features_vector = [
            [
                [
                    self.features_map_to_index[index][value]
                    if index in self.features_map_to_index and value in self.features_map_to_index[index]
                    else 0
                    for index, value in enumerate(value_list)
                    if index in self.features_indices
                ]
                for value_list in document
            ]
            for document in X
        ]

        features_count = len(self.features_indices)

        if extend:
            for out in features_vector:
                out.append([0] * features_count)

        features_vector_padded, _ = pad_sequences(features_vector, [0] * features_count)
        output = np.asarray(features_vector_padded)

        return output

    def empty_features_vector(self) -> Iterable[int]:
        features_count = len(self.features_indices)
        return [0] * features_count


class BERTPreprocessor(object):
    """
    Generic transformer preprocessor for a sequence labelling data set.
    Input are pre-tokenized texts, possibly with features and labels to re-align with the sub-tokenization.
    Rely on HuggingFace Tokenizer library and the tokenizer instance obtained with AutoTokenizer.
    """

    def __init__(self, tokenizer, empty_features_vector=None, empty_char_vector=None):
        self.tokenizer = tokenizer
        self.empty_features_vector = empty_features_vector
        self.empty_char_vector = empty_char_vector

        tokenizer_name = type(self.tokenizer).__name__

        # this flag is used to indicate if the tokenizer is a BPE or sentence piece tokenizer
        self.is_BPE_SP = self.infer_BPE_SP_from_tokenizer_name(tokenizer_name)

    def infer_BPE_SP_from_tokenizer_name(self, tokenizer_name):
        """
        Return true if the tokenizer is a BPE using encoded leading character space with byte-based tokens as in the
        usual sentencepiece tokenizers
        """
        # as we rely on the HuggingFace transformer and tokenizer libraries, we list the BPE tokenizer from the model
        # name - this list was created on January 2023
        bpe_sp_tokenizers = [
            "roberta",
            "gpt2",
            "albert",
            "xlnet",
            "marian",
            "t5",
            "camembert",
            "bart",
            "bigbird",
            "blenderbot",
            "clip",
            "flaubert",
            "fsmt",
            "xlm",
            "longformer",
            "marian",
            "phobert",
            "reformer",
            "rembert",
        ]
        tokenizer_name = tokenizer_name.lower()
        for bpe_tok in bpe_sp_tokenizers:
            if tokenizer_name.find(bpe_tok) != -1:
                return True

        return type(self.tokenizer.backend_tokenizer.model).__name__ == "BPE"

    def tokenize_and_align_features_and_labels(self, texts, chars, text_features, text_labels, maxlen=512):
        """
        Training/evaluation usage with features: sub-tokenize+convert to ids/mask/segments input texts, realign labels
        and features given new tokens introduced by the wordpiece sub-tokenizer.
        texts is a list of texts already pre-tokenized
        """
        target_ids = []
        target_type_ids = []
        target_attention_mask = []
        input_tokens = []
        target_chars = []

        target_features = None
        if text_features is not None:
            target_features = []

        target_labels = None
        if text_labels is not None:
            target_labels = []

        for i, text in enumerate(texts):
            local_chars = chars[i]

            features = None
            if text_features is not None:
                features = text_features[i]

            label_list = None
            if text_labels is not None:
                label_list = text_labels[i]

            input_ids, token_type_ids, attention_mask, chars_block, feature_blocks, target_tags, tokens = (
                self.convert_single_text(text, local_chars, features, label_list, maxlen)
            )
            target_ids.append(input_ids)
            target_type_ids.append(token_type_ids)
            target_attention_mask.append(attention_mask)
            input_tokens.append(tokens)
            target_chars.append(chars_block)

            if target_features is not None:
                target_features.append(feature_blocks)

            if target_labels is not None:
                target_labels.append(target_tags)

        return (
            target_ids,
            target_type_ids,
            target_attention_mask,
            target_chars,
            target_features,
            target_labels,
            input_tokens,
        )

    def convert_single_text(self, text_tokens, chars_tokens, features_tokens, label_tokens, max_seq_length):
        """
        Converts a single sequence input into a single transformer input format using generic tokenizer
        of the transformers library, align other channel input to the new sub-tokenization
        """
        if label_tokens is None:
            # we create a dummy label list to facilitate
            label_tokens = []
            while len(label_tokens) < len(text_tokens):
                label_tokens.append(0)

        if features_tokens is None:
            # we create a dummy feature list to facilitate
            features_tokens = []
            while len(features_tokens) < len(text_tokens):
                features_tokens.append(self.empty_features_vector)

        if chars_tokens is None:
            # we create a dummy feature list to facilitate
            chars_tokens = []
            while len(chars_tokens) < len(text_tokens):
                chars_tokens.append(self.empty_char_vector)

        # sub-tokenization
        encoded_result = self.tokenizer(
            text_tokens,
            add_special_tokens=True,
            is_split_into_words=True,
            max_length=max_seq_length,
            truncation=True,
            return_offsets_mapping=True,
            padding="max_length",
        )

        input_ids = encoded_result.input_ids
        offsets = encoded_result.offset_mapping
        if "token_type_ids" in encoded_result:
            token_type_ids = encoded_result.token_type_ids
        else:
            token_type_ids = [0] * len(input_ids)
        attention_mask = encoded_result.attention_mask
        label_ids = []
        chars_blocks = []
        feature_blocks = []

        # tricks to support BPE/sentence piece tokenizer like GPT2, roBERTa, CamemBERT, etc. which encode prefixed
        # spaces in the tokens (the encoding symbol for this space varies from one model to another)
        new_input_ids = []
        new_attention_mask = []
        new_token_type_ids = []
        new_offsets = []
        # this is a boolean for flag empty tokens (e.g. "▁" or "Ġ") that some BPE tokenizer produces
        empty_token = False
        for i in range(0, len(input_ids)):
            offset = offsets[i]
            if len(self.tokenizer.decode(input_ids[i])) == 0:
                # if a decoded token has a length of 0, it is typically a space added for sentence piece/camembert/GPT2
                # which happens to be then sometimes a single token for unknown reason when with is_split_into_words=True
                # we need to skip this but also remove it from attention_mask, token_type_ids and offsets to stay
                # in sync - note that this case seems not appearing anymore in recent HuggingFace Tokenizer library update
                empty_token = True
                continue
            elif (
                self.is_BPE_SP
                and self.tokenizer.convert_ids_to_tokens(input_ids[i]) not in self.tokenizer.all_special_tokens
                and offset[0] == 0
                and len(self.tokenizer.convert_ids_to_tokens(input_ids[i])) == 1
                and not empty_token
            ):
                # another trick to support sentence piece tokenizer: sometimes a out of vocabulary
                # character is tokenized as several known bytes, leading to 2 tokens for instance
                # with the second one staring from offset 0 too. In order to align correctly the
                # original string, we need to skip this extra spurious token by looking at it decoded
                # form: if we have a start offset of 0 and no encoding leading space symbol, this must
                # be ignored
                empty_token = False
                continue
            elif (
                self.is_BPE_SP
                and self.tokenizer.convert_ids_to_tokens(input_ids[i]) not in self.tokenizer.all_special_tokens
                and offset[0] == 0
                and not empty_token
                and not (
                    self.tokenizer.convert_ids_to_tokens(input_ids[i]).startswith("Ġ")
                    or self.tokenizer.convert_ids_to_tokens(input_ids[i]).startswith("▁")
                    or self.tokenizer.convert_ids_to_tokens(input_ids[i]).startswith(" ")
                )
            ):
                # HuggingFace Roberta and GPT2 tokenizers uses "Ġ" as leading space encoding, other sentencepiece
                # tokenizers usually use "▁" (U+2581) for this. So this should cover existing HuggingFace tokenizers
                # as on January 2023.
                empty_token = False
                continue
            else:
                # valid token
                empty_token = False
                new_input_ids.append(input_ids[i])
                new_attention_mask.append(attention_mask[i])
                new_token_type_ids.append(token_type_ids[i])
                new_offsets.append(offsets[i])
        input_ids = new_input_ids
        attention_mask = new_attention_mask
        token_type_ids = new_token_type_ids
        offsets = new_offsets

        word_idx = -1
        for i, offset in enumerate(offsets):
            if offset[0] == 0 and offset[1] == 0:
                # this is a special token
                label_ids.append("<PAD>")
                chars_blocks.append(self.empty_char_vector)
                feature_blocks.append(self.empty_features_vector)
            else:
                if offset[0] == 0:
                    word_idx += 1
                    # new token
                    label_ids.append(label_tokens[word_idx])
                    feature_blocks.append(features_tokens[word_idx])
                    chars_blocks.append(chars_tokens[word_idx])
                else:
                    # propagate the data to the new sub-token or
                    # dummy/empty input for sub-tokens
                    label_ids.append("<PAD>")
                    chars_blocks.append(self.empty_char_vector)
                    # 2 possibilities, either empty features for sub-tokens or repeating the
                    # feature vector of the prefix sub-token
                    # feature_blocks.append(self.empty_features_vector)
                    feature_blocks.append(features_tokens[word_idx])

        # Zero-pad up to the sequence length.
        while len(input_ids) < max_seq_length:
            input_ids.append(self.tokenizer.pad_token_id)
            token_type_ids.append(self.tokenizer.pad_token_id)
            attention_mask.append(0)
            label_ids.append("<PAD>")
            chars_blocks.append(self.empty_char_vector)
            feature_blocks.append(self.empty_features_vector)

        assert len(input_ids) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(label_ids) == max_seq_length
        assert len(chars_blocks) == max_seq_length
        assert len(feature_blocks) == max_seq_length

        return input_ids, token_type_ids, attention_mask, chars_blocks, feature_blocks, label_ids, offsets

    def convert_single_text_bert(self, text_tokens, chars_tokens, features_tokens, label_tokens, max_seq_length):
        """
        **Deprecated** We now use Transformers Autotokenizer

        Converts a single sequence input into a single BERT input format and align other channel input to this
        new sub-tokenization

        The original BERT implementation works as follow:

        input:
            tokens: [Jim,Henson,was,a,puppeteer]
            labels: [I-PER,I-PER,O,O,O]
        The BERT tokenization will modify sentence and labels as follow:
            tokens: [Jim,Hen,##son,was,a,puppet,##eer]
            labels: [I-PER,I-PER,X,O,O,O,X]

        Here, we don't introduce the X label for added sub-tokens and simply extend the previous label
        and produce:
            tokens: [Jim,Hen,##son,was,a,puppet,##eer]
            labels: [I-PER,I-PER,I-PER,O,O,O,O]

        Notes: input and output labels are text labels, they need to be changed into indices after
        text conversion.
        """

        tokens = []
        # the following is to keep track of additional tokens added by BERT tokenizer,
        # only some of them has a prefix ## that allows to identify them downstream in the process
        tokens_marked = []
        labels = []
        features = []
        chars = []

        if label_tokens is None:
            # we create a dummy label list to facilitate
            label_tokens = []
            while len(label_tokens) < len(text_tokens):
                label_tokens.append(0)

        if features_tokens is None:
            # we create a dummy feature list to facilitate
            features_tokens = []
            while len(features_tokens) < len(text_tokens):
                features_tokens.append(self.empty_features_vector)

        if chars_tokens is None:
            # we create a dummy feature list to facilitate
            chars_tokens = []
            while len(chars_tokens) < len(text_tokens):
                chars_tokens.append(self.empty_char_vector)

        for text_token, label_token, chars_token, features_token in zip(
            text_tokens, label_tokens, chars_tokens, features_tokens
        ):
            text_sub_tokens = self.tokenizer.tokenize(text_token, add_special_tokens=False)

            # we mark added sub-tokens with the "##" prefix in order to restore token back correctly,
            # otherwise the BERT tokenizer do not mark them all with this prefix
            # (note: to be checked if it's the same with the non-original BERT tokenizer)
            text_sub_tokens_marked = self.tokenizer.tokenize(text_token, add_special_tokens=False)
            for i in range(len(text_sub_tokens_marked)):
                if i == 0:
                    continue
                tok = text_sub_tokens_marked[i]
                if not tok.startswith("##"):
                    text_sub_tokens_marked[i] = "##" + tok

            label_sub_tokens = [label_token] + [label_token] * (len(text_sub_tokens) - 1)
            chars_sub_tokens = [chars_token] + [chars_token] * (len(text_sub_tokens) - 1)
            feature_sub_tokens = [features_token] + [features_token] * (len(text_sub_tokens) - 1)

            tokens.extend(text_sub_tokens)
            tokens_marked.extend(text_sub_tokens_marked)
            labels.extend(label_sub_tokens)
            features.extend(feature_sub_tokens)
            chars.extend(chars_sub_tokens)

        if len(tokens) >= max_seq_length - 2:
            tokens = tokens[0 : (max_seq_length - 2)]
            tokens_marked = tokens_marked[0 : (max_seq_length - 2)]
            labels = labels[0 : (max_seq_length - 2)]
            features = features[0 : (max_seq_length - 2)]
            chars = chars[0 : (max_seq_length - 2)]

        input_tokens = []
        input_tokens_marked = []
        segment_ids = []
        label_ids = []
        chars_blocks = []
        feature_blocks = []

        # The convention in BERT is:
        # (a) For sequence pairs:
        #   tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        #   type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
        # (b) For single sequences:
        #   tokens:   [CLS] the dog is hairy . [SEP]
        #   type_ids: 0     0   0   0  0     0 0
        #
        # Where "type_ids" are used to indicate whether this is the first sequence
        # or the second sequence.

        input_tokens.append(self.tokenizer.cls_token)
        input_tokens_marked.append(self.tokenizer.cls_token)
        segment_ids.append(0)
        label_ids.append("<PAD>")
        chars_blocks.append(self.empty_char_vector)
        feature_blocks.append(self.empty_features_vector)

        for i, token in enumerate(tokens):
            input_tokens.append(token)
            segment_ids.append(0)
            label_ids.append(labels[i])
            feature_blocks.append(features[i])
            chars_blocks.append(chars[i])

        for token in tokens_marked:
            input_tokens_marked.append(token)

        input_tokens.append(self.tokenizer.sep_token)
        input_tokens_marked.append(self.tokenizer.sep_token)
        segment_ids.append(0)
        label_ids.append("<PAD>")
        chars_blocks.append(self.empty_char_vector)
        feature_blocks.append(self.empty_features_vector)

        input_ids = self.tokenizer.convert_tokens_to_ids(input_tokens)

        # The mask has 1 for real tokens and 0 for padding tokens
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < max_seq_length:
            input_ids.append(self.tokenizer.pad_token_id)
            input_mask.append(self.tokenizer.pad_token_id)
            segment_ids.append(0)
            label_ids.append("<PAD>")
            chars_blocks.append(self.empty_char_vector)
            feature_blocks.append(self.empty_features_vector)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(label_ids) == max_seq_length
        assert len(chars_blocks) == max_seq_length
        assert len(feature_blocks) == max_seq_length

        return input_ids, input_mask, segment_ids, chars_blocks, feature_blocks, label_ids, input_tokens_marked


class Preprocessor(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        padding=True,
        return_lengths=False,
        return_word_embeddings=False,
        return_casing=False,
        return_features=False,
        return_chars=False,
        return_bert_embeddings=False,
        max_char_length=30,
        feature_preprocessor: FeaturesPreprocessor = None,
    ):

        self.padding = padding
        self.return_lengths = return_lengths
        self.return_word_embeddings = return_word_embeddings
        self.return_casing = return_casing
        self.return_features = return_features
        self.return_chars = return_chars
        self.return_bert_embeddings = return_bert_embeddings
        self.vocab_char = None
        self.vocab_tag = None
        self.vocab_case = [k for k, v in case_index.items()]
        self.max_char_length = max_char_length
        self.feature_preprocessor = feature_preprocessor
        self.indice_tag = None

    def fit(self, X, y):
        chars = {PAD: 0, UNK: 1}
        tags = {PAD: 0}

        # if self.return_chars:

        temp_chars = {c for w in set(itertools.chain(*X)) for c in w}

        sorted_chars = sorted(temp_chars)
        sorted_chars_dict = {c: idx + 2 for idx, c in enumerate(sorted_chars)}
        chars = {**chars, **sorted_chars_dict}

        temp_tags = set(itertools.chain(*y))
        sorted_tags = sorted(temp_tags)
        sorted_tags_dict = {tag: idx + 1 for idx, tag in enumerate(sorted_tags)}
        tags = {**tags, **sorted_tags_dict}

        self.vocab_char = chars
        self.vocab_tag = tags
        self.indice_tag = {i: t for t, i in self.vocab_tag.items()}

        return self

    def extend(self, X, y):
        chars = self.vocab_char
        tags = self.vocab_tag

        temp_chars = {c for w in set(itertools.chain(*X)) for c in w if c not in chars}

        sorted_chars = sorted(temp_chars)
        sorted_chars_dict = {c: idx + len(chars) for idx, c in enumerate(sorted_chars)}
        chars = {**chars, **sorted_chars_dict}

        temp_tags = set(itertools.chain(*y))
        # filter known tags
        temp_tags = {the_tag for the_tag in temp_tags if the_tag not in tags}
        sorted_tags = sorted(temp_tags)
        sorted_tags_dict = {tag: idx + len(tags) for idx, tag in enumerate(sorted_tags)}
        tags = {**tags, **sorted_tags_dict}

        self.vocab_char = chars
        self.vocab_tag = tags

        # refresh tag indices
        self.indice_tag = {i: t for t, i in self.vocab_tag.items()}

        return self

    def transform(self, X, y=None, extend=False, label_indices=False):
        """
        transforms input into sequence
        the optional boolean `extend` indicates that we need to avoid sequence of length 1 alone in a batch
        (which would cause an error with tf)

        Args:
            X: list of list of word tokens
            y: list of list of tags

        Returns:
            numpy array: sentences with char sequences and length
            numpy array: sequence of tags, either one hot encoded (default) or as indices

        if label_indices parameter is true, we encode tags with index integer, otherwise output hot one encoded tags
        """
        chars = []
        lengths = []
        for sent in X:
            char_ids = []
            lengths.append(len(sent))
            for w in sent:
                # if self.return_chars:
                char_ids.append(self.get_char_ids(w))
                if extend:
                    char_ids.append([])

            # if self.return_chars:
            chars.append(char_ids)

        if y is not None:
            pad_index = self.vocab_tag[PAD]
            y = [[self.vocab_tag.get(t, pad_index) for t in sent] for sent in y]
            if extend:
                y[0].append(pad_index)

        if self.padding:
            sents, y = self.pad_sequence(chars, y, label_indices=label_indices)
        else:
            sents = [chars]

        # lengths
        # if self.return_lengths:
        lengths = np.asarray(lengths, dtype=np.int32)
        lengths = lengths.reshape((lengths.shape[0], 1))
        sents.append(lengths)

        return (sents, y) if y is not None else sents

    def fit_features(self, features_batch):
        if self.feature_preprocessor is None:
            return

        return self.feature_preprocessor.fit(features_batch)

    def transform_features(self, features_batch, extend=False):
        return self.feature_preprocessor.transform(features_batch, extend=extend)

    def inverse_transform(self, y):
        """
        send back original label string from label index
        """
        if self.indice_tag is None:
            self.indice_tag = {i: t for t, i in self.vocab_tag.items()}
        return [self.indice_tag[y_] for y_ in y]

    def get_char_ids(self, word):
        return [self.vocab_char.get(c, self.vocab_char[UNK]) for c in word]

    def pad_sequence(self, char_ids, labels=None, label_indices=False):
        """
        pad char and label sequences

        Relatively to labels, if label_indices is True, we encode labels with integer,
        otherwise with hot one encoding
        """
        labels_final = None
        if labels:
            labels_padded, _ = pad_sequences(labels, 0)
            labels_final = np.asarray(labels_padded)
            if not label_indices:
                labels_final = dense_to_one_hot(labels_final, len(self.vocab_tag), nlevels=2)

        # if self.return_chars:
        char_ids, word_lengths = pad_sequences(char_ids, pad_tok=0, nlevels=2, max_char_length=self.max_char_length)
        char_ids = np.asarray(char_ids)
        return [char_ids], labels_final
        # else:
        #    return labels_final

    def empty_features_vector(self) -> Iterable[int]:
        if self.feature_preprocessor is not None:
            return self.feature_preprocessor.empty_features_vector()
        else:
            return None

    def empty_char_vector(self) -> Iterable[int]:
        return [0] * self.max_char_length

    def save(self, file_path):
        variables = vars(self)
        output_dict = {}
        for var in variables.keys():
            if var == "feature_preprocessor" and variables["feature_preprocessor"] is not None:
                output_dict[var] = variables[var].__dict__
            else:
                output_dict[var] = variables[var]

        with open(file_path, "w") as fp:
            json.dump(output_dict, fp, sort_keys=False, indent=4)

    @classmethod
    def load(cls, file_path):
        with open(file_path) as f:
            variables = json.load(f)
            self = cls()
            for key, val in variables.items():
                if key == "feature_preprocessor" and val is not None:
                    preprocessor = FeaturesPreprocessor()
                    preprocessor.__dict__.update(val)
                    if "features_map_to_index" in preprocessor.__dict__:
                        preprocessor.__dict__["features_map_to_index"] = {
                            int(key): val for key, val in preprocessor.__dict__["features_map_to_index"].items()
                        }
                    setattr(self, key, preprocessor)
                else:
                    setattr(self, key, val)

        # fix typing issue for integer keys which become string :/
        self.indice_tag = {i: t for t, i in self.vocab_tag.items()}
        return self


def _pad_sequences(sequences, pad_tok, max_length):
    """
    Args:
        sequences: a generator of list or tuple.
        pad_tok: the char to pad with.

    Returns:
        a list of list where each sublist has same length.
    """
    sequence_padded, sequence_length = [], []

    for seq in sequences:
        seq = list(seq)
        seq_ = seq[:max_length] + [pad_tok] * max(max_length - len(seq), 0)
        sequence_padded += [seq_]
        sequence_length += [min(len(seq), max_length)]

    return sequence_padded, sequence_length


def pad_sequences(sequences, pad_tok=0, nlevels=1, max_char_length=30):
    """
    Args:
        sequences: a generator of list or tuple.
        pad_tok: the char to pad with.

    Returns:
        a list of list where each sublist has same length.
    """
    if nlevels == 1:
        max_length = len(max(sequences, key=len))
        sequence_padded, sequence_length = _pad_sequences(sequences, pad_tok, max_length)
    elif nlevels == 2:
        max_length_word = max_char_length
        sequence_padded, sequence_length = [], []
        for seq in sequences:
            # all words are same length now
            sp, sl = _pad_sequences(seq, pad_tok, max_length_word)
            sequence_padded += [sp]
            sequence_length += [sl]

        max_length_sentence = max(map(lambda x: len(x), sequences))
        sequence_padded, _ = _pad_sequences(sequence_padded, [pad_tok] * max_length_word, max_length_sentence)
        sequence_length, _ = _pad_sequences(sequence_length, 0, max_length_sentence)
    else:
        raise ValueError("nlevels can take 1 or 2, not take {}.".format(nlevels))

    return sequence_padded, sequence_length


def dense_to_one_hot(labels_dense, num_classes, nlevels=1):
    """
    Convert class labels from scalars to one-hot vectors

    Args:
        labels_dense: a dense set of values
        num_classes: the number of classes in output
        nlevels: the number of levels (??)
    """
    if nlevels == 1:
        num_labels = labels_dense.shape[0]
        index_offset = np.arange(num_labels) * num_classes
        labels_one_hot = np.zeros((num_labels, num_classes), dtype=np.int32)
        labels_one_hot.flat[index_offset + labels_dense.ravel()] = 1
        return labels_one_hot
    elif nlevels == 2:
        # assume that labels_dense has same column length
        num_labels = labels_dense.shape[0]
        num_length = labels_dense.shape[1]
        labels_one_hot = np.zeros((num_labels, num_length, num_classes), dtype=np.int32)
        layer_idx = np.arange(num_labels).reshape(num_labels, 1)
        # this index selects each component separately
        component_idx = np.tile(np.arange(num_length), (num_labels, 1))
        # then we use `a` to select indices according to category label
        labels_one_hot[layer_idx, component_idx, labels_dense] = 1
        return labels_one_hot
    else:
        raise ValueError("nlevels can take 1 or 2, not take {}.".format(nlevels))


def prepare_preprocessor(X, y, model_config, features: np.array = None):
    """
    Prepare the preprocessor. If features are passed, configure the feature preprocessor

    From: https://github.com/elifesciences/sciencebeam-trainer-delft/blob/5ceb89bdb9ae56c7f60d68b3aeb7e04dc34cd2be/sciencebeam_trainer_delft/sequence_labelling/preprocess.py#L81
    """
    feature_preprocessor = None
    if features is not None and str.endswith(model_config.architecture, "FEATURES"):
        feature_preprocessor = FeaturesPreprocessor(
            features_indices=model_config.features_indices,
            features_vocabulary_size=model_config.features_vocabulary_size,
        )

    preprocessor = Preprocessor(max_char_length=model_config.max_char_length, feature_preprocessor=feature_preprocessor)
    preprocessor.fit(X, y)

    # Compute features and store information in the model config
    if feature_preprocessor is not None:
        preprocessor.fit_features(features)
        model_config.features_indices = preprocessor.feature_preprocessor.features_indices
        model_config.features_map_to_index = preprocessor.feature_preprocessor.features_map_to_index

    return preprocessor


def to_vector_single(tokens, embeddings, maxlen, lowercase=False, num_norm=True):
    """
    Given a list of tokens convert it to a sequence of word embedding
    vectors with the provided embeddings, introducing <PAD> and <UNK> padding token
    vector when appropriate
    """
    window = tokens[-maxlen:]

    # TBD: use better initializers (uniform, etc.)
    x = np.zeros(
        (maxlen, embeddings.embed_size),
    )

    # TBD: padding should be left and which vector do we use for padding?
    # and what about masking padding later for RNN?
    for i, word in enumerate(window):
        if lowercase:
            word = _lower(word)
        if num_norm:
            word = _normalize_num(word)
        x[i, :] = embeddings.get_word_vector(word).astype("float32")

    return x


def get_subtokens(tokens, maxlen, extend=False, lowercase=False):
    """
    Extract the token list and eventually lowercase or truncate longest sequences

    :param tokens: input tokens
    :param maxlen: maximum length for each sub_token
    :param extend: when set to true, sub tokens will be padded with an additional element
    :param lowercase: when set to true the sub_tokens will be lowercased
    :return:
    """
    subtokens = []
    for i in range(0, len(tokens)):
        local_tokens = []
        for j in range(0, min(len(tokens[i]), maxlen)):
            if lowercase:
                local_tokens.append(_lower(tokens[i][j]))
            else:
                local_tokens.append(tokens[i][j])
        if extend:
            local_tokens.append(UNK)
        subtokens.append(local_tokens)
    return subtokens


def to_casing_single(tokens, maxlen):
    """
    Given a list of tokens set the casing, introducing <PAD> and <UNK> padding
    when appropriate
    """
    window = tokens[-maxlen:]

    # TBD: use better initializers (uniform, etc.)
    x = np.zeros(
        (maxlen),
    )

    # TBD: padding should be left and which vector do we use for padding?
    # and what about masking padding later for RNN?
    for i, word in enumerate(window):
        x[i] = float(_casing(word))

    return x


def _casing(word):
    casing = "other"

    numDigits = 0
    for char in word:
        if char.isdigit():
            numDigits += 1
    digitFraction = numDigits / float(len(word))

    if word.isdigit():
        casing = "numeric"
    elif digitFraction > 0.5:
        casing = "mainly_numeric"
    elif word.islower():
        casing = "allLower"
    elif word.isupper():
        casing = "allUpper"
    elif word[0].isupper():
        casing = "initialUpper"
    elif numDigits > 0:
        casing = "contains_digit"

    return case_index[casing]


def _lower(word):
    return word.lower()


def _normalize_num(word):
    return re.sub(r"[0-9０１２３４５６７８９]", r"0", word)



import tensorflow as tf
import tf_keras

from delft.utilities.crf_wrapper_default import CRFModelWrapperDefault

"""
Alternative CRF model wrapper for models having a BERT/transformer layer.
Loss is modified to ignore labels corresponding to tokens being special transformer symbols (e.g. SEP,
CL, PAD, ...), but also those introduced sub-tokens.
The goal is to have similar effect as in pytorch when using the -100 to ignore some labels when calculating
the loss, but less hacky.
"""


@tf_keras.utils.register_keras_serializable(package="Addons")
class CRFModelWrapperForBERT(CRFModelWrapperDefault):
    def train_step(self, data):
        x, y, sample_weight = self.unpack_training_data(data)

        with tf.GradientTape() as tape:
            (potentials, sequence_length, kernel), decoded_sequence, *_ = self(
                x, training=True, return_crf_internal=True
            )

            mask_value = 0
            # create a boolean mask with False for labels to be ignored
            special_mask = tf.not_equal(y, mask_value)
            special_mask = tf.cast(special_mask, tf.float32)
            # tf.print(special_mask)
            # tf.print(potentials)
            # apply the mask to prediction vectors, it will put to 0
            # weights for label to ignore, normally neutralizing them for
            # loss calculation
            potentials = tf.multiply(potentials, tf.expand_dims(special_mask, -1))
            # tf.print(potentials)

            # experiment: replace 0 by -100 because it's log-based potential
            # the_minus = tf.fill(tf.shape(potentials), -100.0)
            # potentials = tf.where(tf.equal(potentials, 0.0), the_minus, potentials)

            crf_loss = self.compute_crf_loss(potentials, sequence_length, kernel, y, sample_weight)
            loss = crf_loss + tf.reduce_sum(self.losses)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))

        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, decoded_sequence)
        # Return a dict mapping metric names to current value
        orig_results = {m.name: m.result() for m in self.metrics}
        crf_results = {"loss": loss, "crf_loss": crf_loss}
        # crf_results = {"crf_loss": crf_loss}
        return {**orig_results, **crf_results}

    def test_step(self, data):
        x, y, sample_weight = self.unpack_training_data(data)
        (potentials, sequence_length, kernel), decode_sequence, *_ = self(x, training=False, return_crf_internal=True)

        mask_value = 0
        special_mask = tf.not_equal(y, mask_value)
        special_mask = tf.cast(special_mask, tf.float32)
        potentials = tf.multiply(potentials, tf.expand_dims(special_mask, -1))

        crf_loss = self.compute_crf_loss(potentials, sequence_length, kernel, y, sample_weight)
        crf_loss + tf.reduce_sum(self.losses)
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, decode_sequence)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        # results.update({"loss": loss, "crf_loss": crf_loss})  # append loss
        results.update({"crf_loss": crf_loss})  # append loss
        return results



"""
Convert LMDB embeddings from pickled numpy format to raw float32 bytes.

This creates a Java-compatible LMDB database where values are raw float32 arrays.

Usage:
    python -m delft.utilities.convert_lmdb_embeddings \
        --input data/db/glove-840B \
        --output data/db/glove-840B-raw
"""

import argparse
import os
import pickle

import lmdb
import numpy as np
from tqdm import tqdm


def convert_lmdb(input_path: str, output_path: str, batch_size: int = 10000):
    """
    Convert LMDB from pickled numpy arrays to raw float32 bytes.

    Args:
        input_path: Path to source LMDB (with pickled numpy)
        output_path: Path to destination LMDB (with raw float32)
        batch_size: Number of entries to write per transaction
    """
    # Open source database
    src_env = lmdb.open(input_path, readonly=True, max_dbs=1, lock=False)

    # Get stats
    with src_env.begin() as txn:
        stats = txn.stat()
        total_entries = stats["entries"]

    print(f"Source database: {input_path}")
    print(f"  Total entries: {total_entries}")

    # Check one entry to get embedding size
    with src_env.begin() as txn:
        cursor = txn.cursor()
        cursor.first()
        key, value = cursor.item()
        arr = pickle.loads(value)
        embed_size = arr.shape[0]
        print(f"  Embedding size: {embed_size}")
        print(f"  Source format: pickled numpy {arr.dtype}")

    # Create destination database
    os.makedirs(output_path, exist_ok=True)

    # Calculate map size (word bytes + embed floats) per entry, with overhead
    map_size = total_entries * (100 + embed_size * 4) * 2  # 2x overhead
    map_size = max(map_size, 10 * 1024 * 1024 * 1024)  # At least 10GB

    dst_env = lmdb.open(output_path, map_size=map_size, max_dbs=1)

    print(f"\nDestination database: {output_path}")
    print(f"  Map size: {map_size / (1024**3):.1f} GB")
    print(f"  Target format: raw float32 ({embed_size * 4} bytes per vector)")

    # Convert entries
    print(f"\nConverting {total_entries} entries...")

    count = 0
    batch = []

    with src_env.begin() as src_txn:
        cursor = src_txn.cursor()

        for key, value in tqdm(cursor, total=total_entries, desc="Converting"):
            # Unpickle numpy array
            arr = pickle.loads(value)

            # Convert to float32 bytes
            raw_bytes = arr.astype(np.float32).tobytes()

            batch.append((key, raw_bytes))

            if len(batch) >= batch_size:
                # Write batch
                with dst_env.begin(write=True) as dst_txn:
                    for k, v in batch:
                        dst_txn.put(k, v)
                count += len(batch)
                batch = []

    # Write remaining
    if batch:
        with dst_env.begin(write=True) as dst_txn:
            for k, v in batch:
                dst_txn.put(k, v)
        count += len(batch)

    # Close databases
    src_env.close()
    dst_env.close()

    print("\nConversion complete!")
    print(f"  Entries converted: {count}")

    # Verify
    print("\nVerifying...")
    verify_env = lmdb.open(output_path, readonly=True)
    with verify_env.begin() as txn:
        # Check a known word
        test_words = [b"the", b"April", b"2001", b"January"]
        for word in test_words:
            value = txn.get(word)
            if value:
                arr = np.frombuffer(value, dtype=np.float32)
                print(f"  '{word.decode()}': shape={arr.shape}, first 3=[{arr[0]:.4f}, {arr[1]:.4f}, {arr[2]:.4f}]")
    verify_env.close()

    print("\nDone! Java can now read this database as raw float32 bytes.")


def main():
    parser = argparse.ArgumentParser(description="Convert LMDB embeddings from pickled numpy to raw float32")
    parser.add_argument("--input", required=True, help="Path to source LMDB database")
    parser.add_argument("--output", required=True, help="Path to destination LMDB database")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=10000,
        help="Batch size for writing (default: 10000)",
    )

    args = parser.parse_args()

    convert_lmdb(args.input, args.output, args.batch_size)


if __name__ == "__main__":
    main()



import argparse
import json
import os
import time

from delft.sequenceLabelling import Sequence
from delft.sequenceLabelling.reader import load_data_and_labels_xml_file
from delft.utilities.Utilities import t_or_f


def configure(architecture, embeddings_name, batch_size=-1, max_epoch=-1, early_stop=None):

    maxlen = 300
    patience = 5
    o_early_stop = True
    if max_epoch == -1:
        max_epoch = 50

    if batch_size == -1:
        batch_size = 20

    # default bert model parameters
    if architecture.find("BERT") != -1:
        if batch_size == -1:
            batch_size = 10
        o_early_stop = False
        if max_epoch == -1:
            max_epoch = 3

        embeddings_name = None

    if early_stop is not None:
        o_early_stop = early_stop

    return batch_size, maxlen, patience, o_early_stop, max_epoch, embeddings_name


def train(
    embeddings_name=None,
    architecture="BidLSTM_CRF",
    transformer=None,
    learning_rate=None,
    batch_size=-1,
    max_epoch=-1,
    early_stop=None,
    multi_gpu=False,
):
    batch_size, maxlen, patience, early_stop, max_epoch, embeddings_name = configure(
        architecture, embeddings_name, batch_size, max_epoch, early_stop
    )
    root = "data/sequenceLabelling/toxic/"

    train_path = os.path.join(root, "corrected.xml")
    valid_path = os.path.join(root, "valid.xml")

    print("Loading data...")
    x_train, y_train = load_data_and_labels_xml_file(train_path)
    x_valid, y_valid = load_data_and_labels_xml_file(valid_path)
    print(len(x_train), "train sequences")
    print(len(x_valid), "validation sequences")

    model_name = "insult-" + architecture

    model = Sequence(
        model_name,
        max_epoch=max_epoch,
        batch_size=batch_size,
        max_sequence_length=maxlen,
        embeddings_name=embeddings_name,
        architecture=architecture,
        patience=patience,
        early_stop=early_stop,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )
    model.train(x_train, y_train, x_valid=x_valid, y_valid=y_valid, multi_gpu=multi_gpu)
    print("training done")

    # saving the model (must be called after eval for multiple fold training)
    model.save()


# annotate a list of texts, provides results in a list of offset mentions
def annotate(texts, output_format, architecture="BidLSTM_CRF", transformer=None, multi_gpu=False):
    annotations = []

    model_name = "insult-" + architecture

    # load model
    model = Sequence(model_name, architecture=architecture, transformer_name=transformer)
    model.load()

    start_time = time.time()

    annotations = model.tag(texts, output_format, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)

    if output_format == "json":
        annotations["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))
    return annotations


if __name__ == "__main__":
    architectures_word_embeddings = [
        "BidLSTM",
        "BidLSTM_CRF",
        "BidLSTM_ChainCRF",
        "BidLSTM_CNN_CRF",
        "BidLSTM_CNN_CRF",
        "BidGRU_CRF",
        "BidLSTM_CNN",
        "BidLSTM_CRF_CASING",
    ]

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]

    architectures_transformers_based = [
        "BERT",
        "BERT_CRF",
        "BERT_ChainCRF",
        "BERT_CRF_FEATURES",
        "BERT_CRF_CHAR",
        "BERT_CRF_CHAR_FEATURES",
    ]

    architectures = architectures_word_embeddings + architectures_transformers_based

    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]
    parser = argparse.ArgumentParser(
        description="Experimental insult recognizer for the Wikipedia toxic comments dataset"
    )

    parser.add_argument("action")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture",
        default="BidLSTM_CRF",
        help="Type of model architecture to be used, one of " + str(architectures),
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )
    parser.add_argument("--learning-rate", type=float, default=None, help="Initial learning rate")
    parser.add_argument("--max-epoch", type=int, default=-1, help="Maximum number of epochs.")
    parser.add_argument("--batch-size", type=int, default=-1, help="batch-size parameter to be used.")
    parser.add_argument(
        "--early-stop",
        type=t_or_f,
        default=None,
        help="Force early training termination when metrics scores are not improving "
        + "after a number of epochs equals to the patience parameter.",
    )

    parser.add_argument(
        "--multi-gpu",
        default=False,
        help="Enable the support for distributed computing (the batch size needs to be set accordingly using --batch-size)",
        action="store_true",
    )

    args = parser.parse_args()

    if args.action not in ("train", "tag"):
        print("action not specified, must be one of [train,tag]")

    embeddings_name = args.embedding
    architecture = args.architecture
    transformer = args.transformer
    learning_rate = args.learning_rate

    batch_size = args.batch_size
    max_epoch = args.max_epoch
    early_stop = args.early_stop
    multi_gpu = args.multi_gpu

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        train(
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            learning_rate=learning_rate,
            batch_size=batch_size,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
        )

    if args.action == "tag":
        someTexts = [
            "This is a gentle test.",
            "you're a moronic wimp who is too lazy to do research! die in hell !!",
            "This is a fucking test.",
        ]
        result = annotate(someTexts, "json", architecture=architecture, transformer=transformer)
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



import numpy as np
import tf_keras as keras

from delft.sequenceLabelling.preprocess import BERTPreprocessor, Preprocessor, to_casing_single, to_vector_single
from delft.utilities.numpy import shuffle_triple_with_view
from delft.utilities.Tokenizer import tokenizeAndFilterSimple
from delft.utilities.Utilities import len_until_first_pad, truncate_batch_values


class BaseGenerator(keras.utils.Sequence):
    """
    Abstract class for data generator.

    Generate batch of data to feed sequence labeling model, both for training and prediction.

    This generator is for input based on word embeddings. We keep embeddings application outside the
    model to make it considerably more compact and avoid duplication of embeddings layers.
    """

    def __init__(
        self,
        x,
        y,
        batch_size=24,
        preprocessor: Preprocessor = None,
        bert_preprocessor: BERTPreprocessor = None,
        char_embed_size=25,
        embeddings=None,
        max_sequence_length=None,
        tokenize: bool = False,
        shuffle: bool = True,
        features=None,
        output_input_offsets: bool = False,
        use_chain_crf: bool = False,
    ):
        # self.x and self.y are shuffled view of self.original_x and self.original_y
        self.original_x = self.x = x
        self.original_y = self.y = y
        # features here are optional additional features provided in the case of GROBID input for instance
        self.original_features = self.features = features
        self.preprocessor = preprocessor
        self.bert_preprocessor = bert_preprocessor
        if preprocessor:
            self.labels = preprocessor.vocab_tag
        self.batch_size = batch_size
        self.embeddings = embeddings
        self.char_embed_size = char_embed_size
        self.shuffle = shuffle
        self.tokenize = tokenize
        self.max_sequence_length = max_sequence_length
        self.output_input_offsets = output_input_offsets
        self.use_chain_crf = use_chain_crf

    def __len__(self):
        """
        Give the number of batches per epoch
        """
        # The number of batches is set so that each training sample is seen at most once per epoch
        if self.original_x is None:
            return 0
        elif (len(self.original_x) % self.batch_size) == 0:
            return int(np.floor(len(self.original_x) / self.batch_size))
        else:
            return int(np.floor(len(self.original_x) / self.batch_size) + 1)

    @property
    def __getitem__(self, index):
        """
        Generate one batch of data
        """
        raise NotImplementedError("Subclasses should implement this")

    def on_epoch_end(self):
        """
        In case we are training, we can shuffle the training data for the next epoch.
        """
        # If we are predicting, we don't need to shuffle
        if self.original_y is None:
            return

        # shuffle dataset at each epoch
        if self.shuffle:
            self.x, self.y, self.features = shuffle_triple_with_view(
                self.original_x, self.original_y, self.original_features
            )

    @property
    def __data_generation(self, index):
        """
        Generates data containing batch_size samples
        """
        raise NotImplementedError("Subclasses should implement this")


class DataGenerator(BaseGenerator):
    """
    This generator is for input based on word embeddings. We keep embeddings application outside the
    model to make it considerably more compact and avoid duplication of embeddings layers.
    """

    def __init__(
        self,
        x,
        y,
        batch_size=24,
        preprocessor=None,
        bert_preprocessor=None,
        char_embed_size=25,
        embeddings=None,
        max_sequence_length=None,
        tokenize=False,
        shuffle=True,
        features=None,
        output_input_offsets=False,
        use_chain_crf=False,
    ):

        super().__init__(
            x,
            y,
            batch_size=batch_size,
            preprocessor=preprocessor,
            bert_preprocessor=bert_preprocessor,
            char_embed_size=char_embed_size,
            embeddings=embeddings,
            max_sequence_length=max_sequence_length,
            tokenize=tokenize,
            shuffle=shuffle,
            features=features,
            output_input_offsets=output_input_offsets,
            use_chain_crf=use_chain_crf,
        )
        if self.embeddings is not None:
            self.embeddings.reopen_lmdb()
        self.on_epoch_end()

    def __getitem__(self, index):
        """
        Generate one batch of data, batch_l always last input, so that it can be used easily by the training scorer
        """
        batch_x, batch_c, batch_f, batch_a, batch_l, batch_y = self.__data_generation(index)
        if self.preprocessor.return_casing:
            return [batch_x, batch_c, batch_a, batch_l], batch_y
        elif self.preprocessor.return_features:
            return [batch_x, batch_c, batch_f, batch_l], batch_y
        else:
            return [batch_x, batch_c, batch_l], batch_y

    def __data_generation(self, index):
        """
        Generates data containing batch_size samples
        """
        max_iter = min(self.batch_size, len(self.original_x) - self.batch_size * index)

        # restrict data to index window
        sub_x = self.x[(index * self.batch_size) : (index * self.batch_size) + max_iter]

        # tokenize texts in self.x if not already done
        if self.tokenize:
            x_tokenized = [tokenizeAndFilterSimple(text) for text in sub_x]
        else:
            x_tokenized = sub_x

        max_length_f = max_length_x = max((len(tokens) for tokens in x_tokenized))

        if self.max_sequence_length and max_length_x > self.max_sequence_length:
            max_length_x = self.max_sequence_length
            # truncation of sequence at max_sequence_length
            x_tokenized = np.asarray(truncate_batch_values(x_tokenized, self.max_sequence_length), dtype=object)

        # prevent sequence of length 1 alone in a batch (this causes an error in the Chain CRF layer)
        extend = False
        if max_length_x == 1:
            max_length_x += 1
            extend = True

        # generate data
        batch_x = np.zeros((max_iter, max_length_x, self.embeddings.embed_size), dtype="float32")
        for i in range(0, max_iter):
            batch_x[i] = to_vector_single(x_tokenized[i], self.embeddings, max_length_x)

        # store tag embeddings
        batch_y = None
        if self.y is not None:
            # note: tags are always already "tokenized" by input token
            batch_y = self.y[(index * self.batch_size) : (index * self.batch_size) + max_iter]
            max_length_y = max((len(y_row) for y_row in batch_y))

            if self.max_sequence_length and max_length_y > self.max_sequence_length:
                # truncation of sequence at max_sequence_length
                batch_y = np.asarray(truncate_batch_values(batch_y, self.max_sequence_length), dtype=object)

        batch_f = np.zeros((batch_x.shape[0:2]), dtype=np.int32)
        if self.preprocessor.return_features:
            sub_f = self.features[(index * self.batch_size) : (index * self.batch_size) + max_iter]
            if self.max_sequence_length and max_length_f > self.max_sequence_length:
                max_length_f = self.max_sequence_length
                # truncation of sequence at max_sequence_length
                sub_f = truncate_batch_values(sub_f, self.max_sequence_length)
            batch_f = self.preprocessor.transform_features(sub_f, extend=extend)

        batch_a = np.zeros((max_iter, max_length_x), dtype=np.int32)
        if self.preprocessor.return_casing:
            for i in range(0, max_iter):
                batch_a[i] = to_casing_single(x_tokenized[i], max_length_x)

        if self.y is not None:
            if self.use_chain_crf:
                batches, batch_y = self.preprocessor.transform(x_tokenized, batch_y, extend=extend, label_indices=False)
            else:
                batches, batch_y = self.preprocessor.transform(x_tokenized, batch_y, extend=extend, label_indices=True)
            batch_y = np.asarray(truncate_batch_values(batch_y, max_length_x), dtype=np.int32)
        else:
            batches = self.preprocessor.transform(x_tokenized, extend=extend)

        batch_c = np.asarray(batches[0], dtype=np.int32)
        batch_l = batches[1]

        return batch_x, batch_c, batch_f, batch_a, batch_l, batch_y


class DataGeneratorTransformers(BaseGenerator):
    """
    Generate batch of data to feed sequence labeling model, both for training and prediction.

    This generator is for input based on transformer embeddings. We keep embeddings application
    outside the model so that we can serialize the model more easily.
    """

    def __init__(
        self,
        x,
        y,
        batch_size=24,
        preprocessor=None,
        bert_preprocessor=None,
        char_embed_size=25,
        embeddings=None,
        max_sequence_length=None,
        tokenize=False,
        shuffle=True,
        features=None,
        output_input_offsets=False,
        use_chain_crf=False,
    ):

        super().__init__(
            x,
            y,
            batch_size=batch_size,
            preprocessor=preprocessor,
            bert_preprocessor=bert_preprocessor,
            char_embed_size=char_embed_size,
            embeddings=embeddings,
            max_sequence_length=max_sequence_length,
            tokenize=tokenize,
            shuffle=shuffle,
            features=features,
            output_input_offsets=output_input_offsets,
            use_chain_crf=use_chain_crf,
        )

        if self.bert_preprocessor.empty_features_vector is None:
            self.bert_preprocessor.empty_features_vector = self.preprocessor.empty_features_vector()

        if self.embeddings is not None:
            self.embeddings.reopen_lmdb()

        self.on_epoch_end()

    def __getitem__(self, index):
        """
        Generate one batch of data. These data are the input of the models but can also be used by the training scorer
        """
        batch_x, batch_x_types, batch_x_masks, batch_c, batch_f, batch_l, batch_input_offsets, batch_y = (
            self.__data_generation(index)
        )

        # careful with the order of data arrays, double-check the models input as defined in models.py before
        # modifying this

        return_data = [batch_x]

        if self.preprocessor.return_chars:
            return_data += [batch_c]

        if self.preprocessor.return_features:
            return_data += [batch_f]

        return_data += [batch_x_types]

        return_data += [batch_x_masks]

        if self.output_input_offsets:
            # always last input, used when prediction are done to be able to restore the correct labeled sequence
            # this is never routed directly to a model input
            return_data += [batch_input_offsets]

        return return_data, batch_y

    def __data_generation(self, index):
        """
        Generates data containing batch_size samples
        """
        max_iter = min(self.batch_size, len(self.original_x) - self.batch_size * index)

        # restrict data to index window
        sub_x = self.x[(index * self.batch_size) : (index * self.batch_size) + max_iter]

        # tokenize texts in self.x if not already done
        if self.tokenize:
            x_tokenized = [tokenizeAndFilterSimple(text) for text in sub_x]
        else:
            x_tokenized = sub_x

        max_length_f = max_length_x = max((len(tokens) for tokens in x_tokenized))

        if self.max_sequence_length and max_length_x > self.max_sequence_length:
            max_length_x = self.max_sequence_length
            # truncation of sequence at max_sequence_length
            x_tokenized = truncate_batch_values(x_tokenized, self.max_sequence_length)

        # generate data
        batch_y = None

        # tag embeddings
        if self.y is not None:
            # note: tags are always already "tokenized" by input token
            batch_y = self.y[(index * self.batch_size) : (index * self.batch_size) + max_iter]
            max_length_y = max((len(y_row) for y_row in batch_y))

            if self.max_sequence_length and max_length_y > self.max_sequence_length:
                # truncation of sequence at max_sequence_length
                batch_y = np.asarray(truncate_batch_values(batch_y, self.max_sequence_length), dtype=object)

        # features
        if self.preprocessor.return_features:
            sub_f = self.features[(index * self.batch_size) : (index * self.batch_size) + max_iter]
            if self.max_sequence_length and max_length_f > self.max_sequence_length:
                max_length_f = self.max_sequence_length
                # truncation of sequence at max_sequence_length
                sub_f = truncate_batch_values(sub_f, self.max_sequence_length)
            sub_f = self.preprocessor.transform_features(sub_f)
        else:
            sub_f = None

        # chars and length
        batches = self.preprocessor.transform(x_tokenized)
        batch_c = batches[0]
        batch_l = batches[1]

        # to have input as sentence piece token index for transformer layer
        input_ids, token_type_ids, attention_mask, input_chars, input_features, input_labels, input_offsets = (
            self.bert_preprocessor.tokenize_and_align_features_and_labels(
                x_tokenized, batch_c, sub_f, batch_y, maxlen=self.max_sequence_length
            )
        )

        # truncate the batch input vectors for the max length in batch after sub-tokenization
        max_length_x = max((len_until_first_pad(tokens, 0) for tokens in input_ids))

        batch_x = np.asarray(truncate_batch_values(input_ids, max_length_x), dtype=np.int32)
        batch_x_types = np.asarray(truncate_batch_values(token_type_ids, max_length_x), dtype=np.int32)
        batch_x_masks = np.asarray(truncate_batch_values(attention_mask, max_length_x), dtype=np.int32)
        batch_c = np.asarray(truncate_batch_values(input_chars, max_length_x), dtype=np.int32)
        batch_input_offsets = np.asarray(truncate_batch_values(input_offsets, max_length_x), dtype=object)

        if self.y is not None:
            __, batch_y = self.preprocessor.transform(x_tokenized, input_labels, label_indices=True)
            batch_y = np.asarray(truncate_batch_values(batch_y, max_length_x), dtype=np.int32)

        if self.preprocessor.return_features:
            batch_f = np.asarray(truncate_batch_values(input_features, max_length_x), dtype=np.int32)
        else:
            batch_f = np.zeros((batch_x.shape[0:2]), dtype=np.int32)

        return batch_x, batch_x_types, batch_x_masks, batch_c, batch_f, batch_l, batch_input_offsets, batch_y



import numpy as np
import tf_keras as keras

from delft.textClassification.preprocess import create_batch_input_bert, to_vector_single
from delft.utilities.numpy import shuffle_triple_with_view


class DataGenerator(keras.utils.Sequence):
    """
    Generate batch of data to feed text classification model, both for training and prediction.
    For Keras input based on word embeddings, we keep embeddings application outside the model
    to make it considerably more compact and avoid duplication of embeddings layers.

    When the Keras input will feed a BERT layer, sentence piece tokenization is kept outside
    the model so that we can serialize the model and have it more compact.
    """

    def __init__(
        self,
        x,
        y,
        batch_size=256,
        maxlen=300,
        list_classes=[],
        embeddings=(),
        shuffle=True,
        bert_data=False,
        transformer_tokenizer=None,
    ):
        self.x = x
        self.y = y
        self.batch_size = batch_size
        self.maxlen = maxlen
        self.embeddings = embeddings
        self.list_classes = list_classes
        self.shuffle = shuffle
        self.bert_data = bert_data
        self.transformer_tokenizer = transformer_tokenizer
        if self.embeddings is not None:
            self.embeddings.reopen_lmdb()
        self.on_epoch_end()

    def __len__(self):
        """
        Give the number of batches per epoch
        """
        # The number of batches is set so that each training sample is seen at most once per epoch
        if self.x is None:
            return 0
        elif (len(self.x) % self.batch_size) == 0:
            return int(np.floor(len(self.x) / self.batch_size))
        else:
            return int(np.floor(len(self.x) / self.batch_size) + 1)

    def __getitem__(self, index):
        """
        Generate one batch of data
        """
        batch_x, batch_y = self.__data_generation(index)
        return batch_x, batch_y

    def on_epoch_end(self):
        """
        In case we are training, we can shuffle the training data for the next epoch.
        """
        # If we are predicting, we don't need to shuffle
        if self.y is None:
            return

        # other shuffle dataset for next epoch
        if self.shuffle:
            self.x, self.y, _ = shuffle_triple_with_view(self.x, self.y)

    def __data_generation(self, index):
        """
        Generates data containing batch_size samples
        """
        max_iter = min(self.batch_size, len(self.x) - self.batch_size * index)

        if not self.bert_data:
            batch_x = np.zeros((max_iter, self.maxlen, self.embeddings.embed_size), dtype="float32")
        batch_y = None
        if self.y is not None:
            batch_y = np.zeros((max_iter, len(self.list_classes)), dtype="float32")

        # Generate data
        if not self.bert_data:
            for i in range(0, max_iter):
                # for input as word embeddings:
                batch_x[i] = to_vector_single(self.x[(index * self.batch_size) + i], self.embeddings, self.maxlen)
        else:
            # for input as sentence piece token index for BERT layer
            input_ids, input_masks, input_segments = create_batch_input_bert(
                self.x[(index * self.batch_size) : (index * self.batch_size) + max_iter],
                maxlen=self.maxlen,
                transformer_tokenizer=self.transformer_tokenizer,
            )
            # we can use only input indices, but could be reconsidered
            batch_x = np.asarray(input_ids, dtype=np.int32)
            # batch_x_masks = np.asarray(input_masks, dtype=np.int32)
            # batch_x_segments = np.asarray(input_segments, dtype=np.int32)

        # classes are numerical, so nothing to vectorize for y
        for i in range(0, max_iter):
            if self.y is not None:
                batch_y[i] = self.y[(index * self.batch_size) + i]

        return batch_x, batch_y



import logging
from pathlib import Path
from unittest.mock import MagicMock

import pytest

# from py._path.local import LocalPath
import tensorflow as tf

# derived from https://github.com/elifesciences/sciencebeam-trainer-delft/tree/develop/tests

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope="session", autouse=True)
def setup_logging():
    logging.root.handlers = []
    logging.basicConfig(level="INFO")
    logging.getLogger("tests").setLevel("DEBUG")
    # logging.getLogger('sciencebeam_trainer_delft').setLevel('DEBUG')


def _backport_assert_called(mock: MagicMock):
    assert mock.called


@pytest.fixture(scope="session", autouse=True)
def patch_magicmock():
    try:
        MagicMock.assert_called
    except AttributeError:
        MagicMock.assert_called = _backport_assert_called


@pytest.fixture
def temp_dir(tmpdir):
    # convert to standard Path
    return Path(str(tmpdir))


@pytest.fixture(scope="session", autouse=True)
def tf_eager_mode():
    try:
        tf.compat.v1.enable_eager_execution()
    except (ValueError, AttributeError) as e:
        LOGGER.debug("failed to switch to eager mode due to %s", e)



from tf_keras import backend as K
from tf_keras import constraints, initializers, regularizers
from tf_keras.engine.topology import Layer


class Attention(Layer):
    def __init__(
        self,
        step_dim,
        W_regularizer=None,
        b_regularizer=None,
        W_constraint=None,
        b_constraint=None,
        bias=True,
        **kwargs,
    ):
        self.supports_masking = True
        self.init = initializers.get("glorot_uniform")

        self.W_regularizer = regularizers.get(W_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)

        self.W_constraint = constraints.get(W_constraint)
        self.b_constraint = constraints.get(b_constraint)

        self.bias = bias
        self.step_dim = step_dim
        self.features_dim = 0
        super(Attention, self).__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3

        self.W = self.add_weight(
            (input_shape[-1],),
            initializer=self.init,
            name="{}_W".format(self.name),
            regularizer=self.W_regularizer,
            constraint=self.W_constraint,
        )
        self.features_dim = input_shape[-1]

        if self.bias:
            self.b = self.add_weight(
                (input_shape[1],),
                initializer="zero",
                name="{}_b".format(self.name),
                regularizer=self.b_regularizer,
                constraint=self.b_constraint,
            )
        else:
            self.b = None

        self.built = True

    def compute_mask(self, input, input_mask=None):
        return None

    def call(self, x, mask=None):
        features_dim = self.features_dim
        step_dim = self.step_dim

        eij = K.reshape(K.dot(K.reshape(x, (-1, features_dim)), K.reshape(self.W, (features_dim, 1))), (-1, step_dim))

        if self.bias:
            eij += self.b

        eij = K.tanh(eij)

        a = K.exp(eij)

        if mask is not None:
            a *= K.cast(mask, K.floatx())

        a /= K.cast(K.sum(a, axis=1, keepdims=True) + K.epsilon(), K.floatx())

        a = K.expand_dims(a)
        weighted_input = x * a
        return K.sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], self.features_dim


def dot_product(x, kernel):
    """
    Wrapper for dot product operation used in the attention layers
    Args:
        x (): input
        kernel (): weights
    Returns:
    """
    return K.squeeze(K.dot(x, K.expand_dims(kernel)), axis=-1)



import gzip
import json
import os
import re
import xml
from xml.sax import make_parser

import numpy as np
from tqdm import tqdm

from delft.utilities.Tokenizer import tokenizeAndFilter, tokenizeAndFilterSimple


class TEIContentHandler(xml.sax.ContentHandler):
    """
    TEI XML SAX handler for reading mixed content within xml text tags
    """

    # local sentence
    tokens = []
    labels = []

    # all sentences of the document
    sents = []
    allLabels = []

    # working variables
    accumulated = ""
    currentLabel = None

    def __init__(self):
        xml.sax.ContentHandler.__init__(self)

    def startElement(self, name, attrs):
        if self.accumulated != "":
            localTokens = tokenizeAndFilterSimple(self.accumulated)
            for token in localTokens:
                self.tokens.append(token)
                self.labels.append("O")
        if name == "TEI" or name == "tei":
            # beginning of a document
            self.tokens = []
            self.labels = []
            self.sents = []
            self.allLabels = []
        if name == "p":
            # beginning of sentence
            self.tokens = []
            self.labels = []
            self.currentLabel = "O"
        if name == "rs":
            # beginning of entity
            if attrs.getLength() != 0:
                if attrs.getValue("type") != "insult" and attrs.getValue("type") != "threat":
                    print("Invalid entity type:", attrs.getValue("type"))
                self.currentLabel = "<" + attrs.getValue("type") + ">"
        self.accumulated = ""

    def endElement(self, name):
        # print("endElement '" + name + "'")
        if name == "p":
            # end of sentence
            if self.accumulated != "":
                localTokens = tokenizeAndFilterSimple(self.accumulated)
                for token in localTokens:
                    self.tokens.append(token)
                    self.labels.append("O")

            self.sents.append(self.tokens)
            self.allLabels.append(self.labels)
            self.tokens = []
            self.labels = []
        if name == "rs":
            # end of entity
            localTokens = tokenizeAndFilterSimple(self.accumulated)
            begin = True
            if self.currentLabel is None:
                self.currentLabel = "O"
            for token in localTokens:
                self.tokens.append(token)
                if begin:
                    self.labels.append("B-" + self.currentLabel)
                    begin = False
                else:
                    self.labels.append("I-" + self.currentLabel)
            self.currentLabel = None
        self.accumulated = ""

    def characters(self, content):
        self.accumulated += content

    def getSents(self):
        return np.asarray(self.sents, dtype=object)

    def getAllLabels(self):
        return np.asarray(self.allLabels, dtype=object)

    def clear(self):  # clear the accumulator for re-use
        self.accumulated = ""


class ENAMEXContentHandler(xml.sax.ContentHandler):
    """
    ENAMEX-style XML SAX handler for reading mixed content within xml text tags
    """

    # local sentence
    tokens = []
    labels = []

    # all sentences of the document
    sents = []
    allLabels = []

    # working variables
    accumulated = ""
    currentLabel = None
    corpus_type = ""

    def __init__(self, corpus_type="lemonde"):
        xml.sax.ContentHandler.__init__(self)
        self.corpus_type = corpus_type

    def translate_fr_labels(self, mainType, subType):
        # default
        labelOutput = "O"

        if mainType.lower() == "company":
            labelOutput = "business"
        elif mainType.lower() == "fictioncharacter":
            labelOutput = "person"
        elif mainType.lower() == "organization":
            if subType.lower() == "institutionalorganization":
                labelOutput = "institution"
            elif subType.lower() == "company":
                labelOutput = "business"
            else:
                labelOutput = "organisation"
        elif mainType.lower() == "person":
            labelOutput = "person"
        elif mainType.lower() == "location":
            labelOutput = "location"
        elif mainType.lower() == "poi":
            labelOutput = "location"
        elif mainType.lower() == "product":
            labelOutput = "artifact"

        return labelOutput

    def startElement(self, name, attrs):
        if self.accumulated != "":
            localTokens = tokenizeAndFilterSimple(self.accumulated)
            for token in localTokens:
                self.tokens.append(token)
                self.labels.append("O")
        if name == "corpus" or name == "DOC":
            # beginning of a document
            self.tokens = []
            self.labels = []
            self.sents = []
            self.allLabels = []
        if name == "sentence":
            # beginning of sentence
            self.tokens = []
            self.labels = []
            self.currentLabel = "O"
        if name == "ENAMEX" or name == "EX_ENAMEX":
            # beginning of entity
            if attrs.getLength() != 0:
                # if attrs.getValue("type") != 'insult' and attrs.getValue("type") != 'threat':
                #    print("Invalid entity type:", attrs.getValue("type"))
                attrs.getNames()
                mainType = None
                if "type" in attrs:
                    mainType = attrs.getValue("type")
                if "TYPE" in attrs:
                    mainType = attrs.getValue("TYPE")
                if mainType is None:
                    print("ENAMEX element without type attribute!")

                if "sub_type" in attrs:
                    subType = attrs.getValue("sub_type")
                else:
                    subType = ""
                if self.corpus_type == "lemonde":
                    self.currentLabel = "<" + self.translate_fr_labels(mainType, subType) + ">"
                else:
                    self.currentLabel = "<" + mainType + ">"
        self.accumulated = ""

    def endElement(self, name):
        # print("endElement '" + name + "'")
        if name == "sentence":
            # end of sentence
            if self.accumulated != "":
                localTokens = tokenizeAndFilterSimple(self.accumulated)
                for token in localTokens:
                    self.tokens.append(token)
                    self.labels.append("O")

            self.sents.append(self.tokens)
            self.allLabels.append(self.labels)
            self.tokens = []
            self.labels = []
        if name == "ENAMEX" or name == "EX_ENAMEX":
            # end of entity
            localTokens = tokenizeAndFilterSimple(self.accumulated)
            begin = True
            if self.currentLabel is None:
                self.currentLabel = "O"
            for token in localTokens:
                self.tokens.append(token)
                if begin:
                    self.labels.append("B-" + self.currentLabel)
                    begin = False
                else:
                    self.labels.append("I-" + self.currentLabel)
            self.currentLabel = None
        self.accumulated = ""

    def characters(self, content):
        self.accumulated += content

    def getSents(self):
        return np.asarray(self.sents)

    def getAllLabels(self):
        return np.asarray(self.allLabels)

    def clear(self):  # clear the accumulator for re-use
        self.accumulated = ""


def load_data_and_labels_xml_string(stringXml):
    """
    Load data and label from a string
    the format is as follow:
    <p>
        bla bla you are a <rs type="insult">CENSURED</rs>,
        and I will <rs type="threat">find and kill</rs> you bla bla
    </p>
    only the insulting expression is labelled, and similarly only the threat
    "action" is tagged

    Returns:
        tuple(numpy array, numpy array): data and labels

    """
    # as we have XML mixed content, we need a real XML parser...
    parser = make_parser()
    handler = TEIContentHandler()
    parser.setContentHandler(handler)
    parser.parseString(stringXml)
    tokens = handler.getSents()
    labels = handler.getAllLabels()
    return tokens, labels


def load_data_and_labels_xml_file(filepathXml):
    """
    Load data and label from an XML file
    the format is as follow:
    <p>
        bla bla you are a <rs type="insult">CENSURED</rs>,
        and I will <rs type="threat">find and kill</rs> you bla bla
    </p>
    only the insulting expression is labelled, and similarly only the threat
    "action" is tagged

    Returns:
        tuple(numpy array, numpy array): data and labels

    """
    # as we have XML mixed content, we need a real XML parser...
    parser = make_parser()
    handler = TEIContentHandler()
    parser.setContentHandler(handler)
    parser.parse(filepathXml)
    tokens = handler.getSents()
    labels = handler.getAllLabels()
    return tokens, labels


def load_data_crf_file(filepath):
    """
    Load data, features and label from a CRF matrix string
    the format is as follow:

    token_0 f0_0 f0_1 ... f0_n label_0
    token_1 f1_0 f1_1 ... f1_n label_1
    ...
    token_m fm_0 fm_1 ... fm_n label_m

    field separator can be either space or tab
    input file can be gzipped or not

    Returns:
        tuple(numpy array, numpy array): tokens, features

    """

    if filepath.endswith(".gz"):
        with gzip.open(filepath, "rt") as f:
            sents, featureSets = load_data_crf_content(f)
    else:
        with open(filepath) as f:
            sents, featureSets = load_data_crf_content(f)

    return np.asarray(sents, dtype=object), np.asarray(featureSets, dtype=object)


def load_data_and_labels_crf_file(filepath):
    """
    Load data, features and label from a CRF matrix file path in CRF++ or Wapiti supported format.
    the format is as follow:

    token_0 f0_0 f0_1 ... f0_n label_0
    token_1 f1_0 f1_1 ... f1_n label_1
    ...
    token_m fm_0 fm_1 ... fm_n label_m

    field separator can be either space or tab
    input file can be gzipped or not

    Returns:
        tuple(numpy array, numpy array, numpy array): tokens, labels, features

    """
    sents = []
    labels = []
    featureSets = []

    if filepath.endswith(".gz"):
        with gzip.open(filepath, "rt") as f:
            sents, labels, featureSets = load_data_and_labels_crf_content(f)
    else:
        with open(filepath) as f:
            sents, labels, featureSets = load_data_and_labels_crf_content(f)

    return np.asarray(sents, dtype=object), np.asarray(labels, dtype=object), np.asarray(featureSets, dtype=object)


def load_data_and_labels_crf_content(the_file):
    """
    Load data, features and labels from a CRF matrix file in CRF++ or Wapiti supported format.

    the format is as follow:

    token_0 f0_0 f0_1 ... f0_n label_0
    token_1 f1_0 f1_1 ... f1_n label_1
    ...
    token_m fm_0 fm_1 ... fm_n label_n

    field separator can be either space or tab

    Returns:
        tuple(numpy array, numpy array, numpy array): tokens, labels, features
    """
    sents = []
    labels = []
    featureSets = []

    tokens, tags, features = [], [], []
    for line in the_file:
        line = line.strip()
        if len(line) == 0:
            if len(tokens) != 0:
                sents.append(tokens)
                labels.append(tags)
                featureSets.append(features)
                tokens, tags, features = [], [], []
        else:
            pieces = re.split(" |\t", line)
            token = pieces[0]
            tag = pieces[len(pieces) - 1]
            localFeatures = pieces[1 : len(pieces) - 1]
            tokens.append(token)
            tags.append(_translate_tags_grobid_to_IOB(tag))
            features.append(localFeatures)

    if len(tokens) == len(tags) == len(features) > 0:
        sents.append(tokens)
        labels.append(tags)
        featureSets.append(features)
        print("Adding the final items from the input file. ")

    assert "Tokens, tags and features haven't got the same size", len(tokens) == len(tags) == len(features)

    return sents, labels, featureSets


def load_data_crf_content(the_file):
    sents = []
    featureSets = []
    localFeatures = []

    tokens, features = [], []
    for line in the_file:
        line = line.strip()
        if len(line) == 0:
            if len(tokens) != 0:
                sents.append(tokens)
                featureSets.append(features)
                tokens, features = [], []
        else:
            pieces = re.split(" |\t", line)
            token = pieces[0]
            localFeatures = pieces[1 : len(pieces)]
            tokens.append(token)
            features.append(localFeatures)

    if len(tokens) == len(features) > 0:
        sents.append(tokens)
        featureSets.append(features)
        print("Adding the final items from the input file. ")

    assert "Tokens, tags and features haven't got the same size", len(tokens) == len(features)

    return sents, featureSets


def load_data_and_labels_crf_string(crfString):
    """
    Load data, features and labels from a CRF matrix string in CRF++ or Wapiti supported format.
    the format is as follow:

    token_0 f0_0 f0_1 ... f0_n label_0
    token_1 f1_0 f1_1 ... f1_n label_1
    ...
    token_m fm_0 fm_1 ... fm_n label_n

    field separator can be either space or tab

    Returns:
        tuple(numpy array, numpy array, numpy array): tokens, labels, features

    """
    sents = []
    labels = []
    featureSets = []
    tokens, tags, features = [], [], []
    for line in crfString.splitlines():
        line = line.strip(" \t")
        if len(line) == 0:
            if len(tokens) != 0:
                sents.append(tokens)
                labels.append(tags)
                featureSets.append(features)
                tokens, tags, features = [], [], []
        else:
            # pieces = line.split('\t')
            pieces = re.split(" |\t", line)
            token = pieces[0]
            tag = pieces[len(pieces) - 1]
            localFeatures = pieces[1 : len(pieces) - 1]
            tokens.append(token)
            tags.append(_translate_tags_grobid_to_IOB(tag))
            features.append(localFeatures)
    # last sequence
    if len(tokens) != 0:
        sents.append(tokens)
        labels.append(tags)
        featureSets.append(features)
    return sents, labels, featureSets


def load_data_crf_string(crfString):
    """
    Load data and features (no label!) from a CRF matrix file in CRF++ or Wapiti supported format.
    the format is as follow:

    token_0 f0_0 f0_1 ... f0_n
    token_1 f1_0 f1_1 ... f1_n
    ...
    token_m fm_0 fm_1 ... fm_n

    field separator can be either space or tab

    Returns:
        tuple(numpy array, numpy array): tokens, features

    """
    sents = []
    featureSets = []
    tokens, features = [], []
    # print("crfString:", crfString)
    for line in crfString.splitlines():
        line = line.strip(" \t")
        if len(line) == 0:
            if len(tokens) != 0:
                sents.append(tokens)
                featureSets.append(features)
                tokens, features = [], []
        else:
            pieces = re.split(" |\t", line)
            token = pieces[0]
            localFeatures = pieces[1 : len(pieces)]
            tokens.append(token)
            features.append(localFeatures)
    # last sequence
    if len(tokens) != 0:
        sents.append(tokens)
        featureSets.append(features)

    # print('sents:', len(sents))
    # print('featureSets:', len(featureSets))
    return sents, featureSets


def _translate_tags_grobid_to_IOB(tag):
    """
    Convert labels as used by GROBID to the more standard IOB2
    """
    if tag.endswith("other>"):
        # outside
        return "O"
    elif tag.startswith("I-"):
        # begin
        return "B-" + tag[2:]
    elif tag.startswith("<"):
        # inside
        return "I-" + tag
    else:
        return tag


def load_data_and_labels_conll(filename):
    """
    Load data and label from a file.

    Args:
        filename (str): path to the file.

        The file format is tab-separated values.
        A blank line is required at the end of a sentence.

        For example:
        ```
        EU  B-ORG
        rejects O
        German  B-MISC
        call	O
        to	O
        boycott	O
        British	B-MISC
        lamb	O
        .	O

        Peter	B-PER
        Blackburn	I-PER
        ...
        ```

    Returns:
        tuple(numpy array, numpy array): data and labels

    """

    # TBD: ideally, for consistency, the tokenization in the CoNLL files should not be enforced,
    # only the standard DeLFT tokenization should be used, in line with the word embeddings
    sents, labels = [], []
    with open(filename, encoding="UTF-8") as f:
        words, tags = [], []
        for line in f:
            line = line.rstrip()
            if len(line) == 0 or line.startswith("-DOCSTART-") or line.startswith("#begin document"):
                if len(words) != 0:
                    sents.append(words)
                    labels.append(tags)
                    words, tags = [], []
            else:
                if len(line.split("\t")) == 2:
                    word, tag = line.split("\t")
                else:
                    word, _, tag = line.split("\t")
                words.append(word)
                tags.append(tag)

    return np.asarray(sents, dtype=object), np.asarray(labels, dtype=object)


def load_data_and_labels_conll_with_document_context(filename, max_context_window=400):
    """
    Load data and label from a file. In this alternative, we do not segment by sentence and
    we keep a maximum of document context according to a context window size.

    Args:
        filename (str): path to the file.

        The file format is tab-separated values.
        A blank line is required at the end of a sentence.

        For example:
        ```
        EU  B-ORG
        rejects O
        German  B-MISC
        call    O
        to  O
        boycott O
        British B-MISC
        lamb    O
        .   O

        Peter   B-PER
        Blackburn   I-PER
        ...
        ```

    Returns:
        tuple(numpy array, numpy array): data and labels

    """

    # TBD: ideally, for consistency, the tokenization in the CoNLL files should not be enforced,
    # only the standard DeLFT tokenization should be used, in line with the word embeddings
    documents, sents, labels = [], [], []
    with open(filename, encoding="UTF-8") as f:
        words, tags = [], []
        for line in f:
            line = line.rstrip()
            if line.startswith("-DOCSTART-") or line.startswith("#begin document"):
                if len(words) != 0:
                    sents.append(words)
                    labels.append(tags)
                    words, tags = [], []
                if len(sents) != 0:
                    documents.append(sents)
                    sents = []
            elif len(line) == 0:
                if len(words) != 0:
                    sents.append(words)
                    labels.append(tags)
                    words, tags = [], []
            else:
                if len(line.split("\t")) == 2:
                    word, tag = line.split("\t")
                else:
                    word, _, tag = line.split("\t")
                words.append(word)
                tags.append(tag)

    # regroup sentences according document contexts following context window

    return np.asarray(sents, dtype=object), np.asarray(labels, dtype=object)


def load_data_and_labels_lemonde(filepathXml):
    """
    Load data and label from Le Monde XML corpus file
    the format is ENAMEX-style, as follow:
    <sentence id="E14">Les ventes de micro-ordinateurs en <ENAMEX type="Location" sub_type="Country"
        eid="2000000003017382" name="Republic of France">France</ENAMEX> se sont ralenties en 1991. </sentence>

    Returns:
        tuple(numpy array, numpy array): data and labels

    """
    # as we have XML mixed content, we need a real XML parser...
    parser = make_parser()
    handler = ENAMEXContentHandler()
    parser.setContentHandler(handler)
    parser.parse(filepathXml)
    tokens = handler.getSents()
    labels = handler.getAllLabels()

    return tokens, labels


def load_data_and_labels_ontonotes(ontonotesRoot, lang="en"):
    """
    Load data and label from Ontonotes 5.0 pseudo-XML corpus files
    the format is ENAMEX-style, as follow, with one sentence per line:
    <doc>
        <ENAMEX TYPE="DATE">Today</ENAMEX> , one newspaper headline warned of civil war .
        ...
    </doc>

    Returns:
        tuple(numpy array, numpy array): data and labels

    """
    # assuming we have the root of ontonotes corpus, we iterate through the sub-directories
    # and process all .name files
    nb_files = 0
    # map lang and subdir names
    if lang == "zh":
        pass
    elif lang == "ar":
        pass

    tokens = []
    labels = []

    # first pass to get number of files
    for subdir, dirs, files in os.walk(ontonotesRoot):
        for file in files:
            if "/english/" in subdir and file.endswith(".name"):
                # remove old/new testament
                if "/pt/" in subdir:
                    continue
                # print(os.path.join(subdir, file))
                nb_files += 1
    nb_total_files = nb_files
    # print(nb_total_files, 'total files')

    nb_files = 0
    pbar = tqdm(total=nb_total_files)
    for subdir, dirs, files in os.walk(ontonotesRoot):
        for file in files:
            if "/english/" in subdir and file.endswith(".name"):
                # remove old/new testament
                if "/pt/" in subdir:
                    continue
                handler = ENAMEXContentHandler(corpus_type="ontonotes")
                # massage a bit the pseudo-XML so that it looks XML
                with open(os.path.join(subdir, file), encoding="UTF-8") as f:
                    content = '<?xml version="1.0" encoding="utf-8"?>\n'
                    for line in f:
                        line = line.strip()
                        if len(line) > 2 and line[-2] == "/":
                            line = line[: len(line) - 2] + line[-1]

                        if len(line) != 0:
                            if "<DOC" not in line and "</DOC" not in line:
                                content += "<sentence>" + line + "</sentence>\n"
                            else:
                                content += line + "\n"
                    # print(content)
                    xml.sax.parseString(content, handler)
                    tokens.extend(handler.sents)
                    labels.extend(handler.allLabels)
                    nb_files += 1
                    pbar.update(1)
    pbar.close()
    print("nb total sentences:", len(tokens))
    total_tokens = 0
    for sentence in tokens:
        total_tokens += len(sentence)
    print("nb total tokens:", total_tokens)

    final_tokens = np.asarray(tokens)
    final_label = np.asarray(labels)

    return final_tokens, final_label


def load_data_and_labels_json_offsets(jsonCorpus, tokenizer=None):
    """
    Load data and labels from json corpus where annotations are expressed with offsets.
    This requires a tokenizer passed as parameter. If tokenizer is None, we use the generic
    Indo-European tokenizer.

    Note: input file can be gzipped or not

    {
    "lang": "en",
    "level": "sentence",
    "documents": [
        {
            "id": "10.1371/journal.pone.0198300",
            "body_text": [
                {
                    "text": "The test was designed so that bacteria were collected at 1 hour and 6 hours after start time on each day of testing.",
                    "annotation_spans": [
                        {
                            "start": 30,
                            "end": 38,
                            "text": "bacteria",
                            "type": "dataset",
                            "datatype": "Tabular Data:Sample Table"
                        }
                    ]
                },
            ]
        }
    }

    Returns:
        tuple(numpy array, numpy array): data and labels
    """
    if not os.path.exists(jsonCorpus):
        print("Invalid path file: ", jsonCorpus)
        return None, None

    all_tokens = []
    all_labels = []

    if jsonCorpus.endswith(".gz"):
        corpus_file = gzip.open(jsonCorpus, "rt")
    else:
        corpus_file = open(jsonCorpus, "rt")

    jsonDocuments = json.load(corpus_file)
    if "documents" in jsonDocuments:
        for jsonDocument in jsonDocuments["documents"]:
            if "body_text" in jsonDocument:
                for text_piece in jsonDocument["body_text"]:
                    if "text" in text_piece:
                        tokens = []
                        labels = []
                        text = text_piece["text"]
                        local_tokens, local_offsets = tokenizeAndFilter(text)
                        spans = []
                        if "annotation_spans" in text_piece:
                            for annotation_span in text_piece["annotation_spans"]:
                                local_type = None
                                if "type" in annotation_span:
                                    local_type = annotation_span["type"]
                                    local_type = local_type.replace(" ", "_")
                                if "text" in annotation_span:
                                    localText = annotation_span["text"]
                                    if text[annotation_span["start"] : annotation_span["end"]] != localText:
                                        print(
                                            "warning - offsets and chunk mismatch: "
                                            + localText
                                            + " / "
                                            + str(annotation_span["start"])
                                            + ":"
                                            + str(annotation_span["end"])
                                        )

                                if local_type is None:
                                    print(
                                        "warning - empty label: "
                                        + localText
                                        + " / "
                                        + str(annotation_span["start"])
                                        + ":"
                                        + str(annotation_span["end"])
                                    )
                                spans.append([annotation_span["start"], annotation_span["end"], local_type])
                        i = 0
                        for local_token in local_tokens:
                            tokens.append(local_token)
                            offset = local_offsets[i]
                            found = False
                            for span in spans:
                                if span[0] <= offset[0] and (offset[1] <= span[1] or offset[0] < span[1]):
                                    if span[0] == offset[0]:
                                        labels.append("B-" + span[2])
                                    else:
                                        labels.append("I-" + span[2])
                                    found = True
                                    break
                            if not found:
                                labels.append("O")
                            i += 1

                        all_tokens.append(tokens)
                        all_labels.append(labels)

    corpus_file.close()

    final_tokens = np.asarray(all_tokens)
    final_labels = np.asarray(all_labels)

    return final_tokens, final_labels


if __name__ == "__main__":
    # some tests
    xmlPath = "../../data/sequenceLabelling/toxic/train.xml"
    print(xmlPath)
    sents, allLabels = load_data_and_labels_xml_file(xmlPath)
    print("toxic tokens:", sents)
    print("toxic labels:", allLabels)

    xmlPath = "../../data/sequenceLabelling/toxic/test.xml"
    print(xmlPath)
    sents, allLabels = load_data_and_labels_xml_file(xmlPath)
    print("toxic tokens:", sents)
    print("toxic labels:", allLabels)

    crfPath = "../../data/sequenceLabelling/grobid/date/date-060518.train"
    print(crfPath)
    x_all, y_all, f_all = load_data_and_labels_crf_file(input)
    print("grobid date tokens:", x_all)
    print("grobid date labels:", y_all)
    print("grobid date features:", f_all)

    with open(crfPath, "r") as theFile:
        string = theFile.read()

    x_all, f_all = load_data_and_labels_crf_string(string)
    print("grobid date tokens:", x_all)
    print("grobid date features:", f_all)



import logging

from delft.sequenceLabelling import reader

LOGGER = logging.getLogger(__name__)


def test_load_data_crf_string():
    x, y = (
        reader.load_data_crf_string("""In in I In In In n In In In INITCAP NODIGIT 0 NOPUNCT In Xx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
just just j ju jus just t st ust just NOCAPS NODIGIT 0 NOPUNCT just xxxx x SAMEFONT SAMEFONTSIZE false false BASELINE false

Soon soon S So Soo Soon n on oon Soon INITCAP NODIGIT 0 NOPUNCT Soon Xxxx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
after after a af aft afte r er ter fter NOCAPS NODIGIT 0 NOPUNCT after xxxx x SAMEFONT SAMEFONTSIZE false false BASELINE false

Therefore therefore T Th The Ther e re ore fore INITCAP NODIGIT 0 NOPUNCT Therefore Xxxx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
, , , , , , , , , , ALLCAPS NODIGIT 1 COMMA , , , SAMEFONT SAMEFONTSIZE false false BASELINE false

By by B By By By y By By By INITCAP NODIGIT 0 NOPUNCT By Xx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
replacing replacing r re rep repl g ng ing cing NOCAPS NODIGIT 0 NOPUNCT replacing xxxx x SAMEFONT SAMEFONTSIZE false false BASELINE false

Meanwhile meanwhile M Me Mea Mean e le ile hile INITCAP NODIGIT 0 NOPUNCT Meanwhile Xxxx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
, , , , , , , , , , ALLCAPS NODIGIT 1 COMMA , , , SAMEFONT SAMEFONTSIZE false false BASELINE false

More more M Mo Mor More e re ore More INITCAP NODIGIT 0 NOPUNCT More Xxxx Xx SAMEFONT SAMEFONTSIZE false false BASELINE false
excitingly excitingly e ex exc exci y ly gly ngly NOCAPS NODIGIT 0 NOPUNCT excitingly xxxx x SAMEFONT SAMEFONTSIZE false false BASELINE false
, , , , , , , , , , ALLCAPS NODIGIT 1 COMMA , , , SAMEFONT SAMEFONTSIZE false false BASELINE false
""")
    )

    assert len(x) == 6
    assert x[0][0] == "In"
    assert x[0][1] == "just"

    assert x[3][1] == "replacing"

    assert x[4][0] == "Meanwhile"



import os

import pytest

from .test_data import TEST_DATA_PATH


@pytest.fixture
def preprocessor1():
    return os.path.join(TEST_DATA_PATH, "preprocessor.json")


@pytest.fixture
def preprocessor2():
    return os.path.join(TEST_DATA_PATH, "preprocessor2.json")



import regex as re

# Generic simple tokenizer for Indo-European languages
# also python side of GROBID default tokenizer, used for Indo-European languages

delimiters = "\n\r\t\f\u00a0([ •*,:;?.!/)-−–‐\"“”‘’'`$]*\u2666\u2665\u2663\u2660\u00a0\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u200b"
regex = "|".join(map(re.escape, delimiters))
pattern = re.compile("(" + regex + ")")
# additional parenthesis above are for capturing delimiters and keep then in the token list

blanks = " \t\n\u00a0\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u200b"


def tokenize(text):
    """
    Tokenization following the above pattern with offset information and keep
    blank characters (space family) as tokens
    """
    return tokenizeAndFilter(text, filterBlank=False)


def tokenizeAndFilter(text, filterBlank=True):
    """
    Tokenization following the above pattern with offset information and with
    filtering of blank characters (space family)
    """
    offset = 0
    offsets = []
    tokens = []
    for index, match in enumerate(pattern.split(text)):
        tokens.append(match)
        position = (offset, offset + len(match))
        offsets.append(position)
        offset = offset + len(match)

    finalTokens = []
    finalOffsets = []
    i = 0
    for token in tokens:
        if not filterBlank or token not in blanks:
            finalTokens.append(token)
            finalOffsets.append(offsets[i])
        i += 1
    return finalTokens, finalOffsets


def tokenizeAndFilterSimple(text):
    """
    Tokenization following the above pattern without offset information
    """
    tokens = []
    for index, match in enumerate(pattern.split(text)):
        tokens.append(match)

    finalTokens = []
    i = 0
    for token in tokens:
        if token not in blanks:
            finalTokens.append(token)
        i += 1

    return finalTokens


def filterSpace(token):
    return token not in blanks



import gzip
import json

import numpy as np
import pandas as pd

from delft.utilities.numpy import shuffle_triple_with_view


def load_texts_and_classes(filepath):
    """
    Load texts and classes from a file in the following simple tab-separated format:

    id_0    text_0  class_00 ...    class_n0
    id_1    text_1  class_01 ...    class_n1
    ...
    id_m    text_m  class_0m  ...   class_nm

    text has no EOF and no tab

    Returns:
        tuple(numpy array, numpy array): texts and classes

    """
    texts = []
    classes = []

    with open(filepath) as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            pieces = line.split("\t")
            if len(pieces) < 3:
                print("Warning: number of fields in the data file too low for line:", line)
            texts.append(pieces[1])
            classes.append(pieces[2:])

    return np.asarray(texts), np.asarray(classes)


def load_texts_and_classes_pandas(filepath):
    """
    Load texts and classes from a file in csv format using pandas dataframe, with format as follow:

    id      text    class_0     ... class_n
    id_0    text_0  class_00    ... class_n0
    id_1    text_1  class_01    ... class_n1
    ...
    id_m    text_m  class_0m    ... class_nm

    It should support any CSV file format.

    Returns:
        tuple(numpy array, numpy array): texts and classes

    """

    df = pd.read_csv(filepath)
    df.iloc[:, 1].fillna("MISSINGVALUE", inplace=True)

    texts_list = []
    for j in range(0, df.shape[0]):
        texts_list.append(df.iloc[j, 1])

    classes = df.iloc[:, 2:]
    classes_list = classes.values.tolist()

    return np.asarray(texts_list), np.asarray(classes_list)


def load_texts_and_classes_pandas_no_id(filepath):
    """
    Load texts and classes from a file in csv format using pandas dataframe, with format as follow:

    text    class_0     ... class_n
    text_0  class_00    ... class_n0
    text_1  class_01    ... class_n1
    ...
    text_m  class_0m    ... class_nm

    It should support any CSV file format.

    Returns:
        tuple(numpy array, numpy array): texts and classes

    """

    df = pd.read_csv(filepath)
    df.iloc[:, 1].fillna("MISSINGVALUE", inplace=True)

    texts_list = []
    for j in range(0, df.shape[0]):
        texts_list.append(df.iloc[j, 0])

    classes = df.iloc[:, 1:]
    classes_list = classes.values.tolist()

    return np.asarray(texts_list), np.asarray(classes_list)


def load_texts_pandas(filepath):
    """
    Load texts from a file in csv format using pandas dataframe:

    id      text
    id_0    text_0
    id_1    text_1
    ...
    id_m    text_m

    It should support any CSV file format.

    Returns:
        numpy array: texts

    """

    df = pd.read_csv(filepath)
    df.iloc[:, 1].fillna("MISSINGVALUE", inplace=True)

    texts_list = []
    for j in range(0, df.shape[0]):
        texts_list.append(df.iloc[j, 1])

    return np.asarray(texts_list)


def load_citation_sentiment_corpus(filepath):
    """
    Load texts from the citation sentiment corpus:

    Source_Paper  Target_Paper    Sentiment   Citation_Text

    sentiment "value" can o (neutral), p (positive), n (negative)

    Returns:
        tuple(numpy array, numpy array): texts and polarity

    """

    texts = []
    polarities = []

    with open(filepath) as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            if line.startswith("#"):
                continue

            pieces = line.split("\t")
            if len(pieces) != 4:
                print("Warning: incorrect number of fields in the data file for line:", line)
                continue
            text = pieces[3]
            # remove start/end quotes
            text = text[1 : len(text) - 1]
            texts.append(text)

            polarity = []
            if pieces[2] == "n":
                polarity.append(1)
            else:
                polarity.append(0)
            if pieces[2] == "o":
                polarity.append(1)
            else:
                polarity.append(0)
            if pieces[2] == "p":
                polarity.append(1)
            else:
                polarity.append(0)
            polarities.append(polarity)

    return np.asarray(texts), np.asarray(polarities)


def load_citation_intent_corpus(filepath):
    """
    Load texts from the citation intent corpus multicite https://github.com/allenai/multicite (NAACL 2022)

    Source_Paper  Target_Paper    Sentiment   Citation_Text

    sentiment "value" can o (neutral), p (positive), n (negative)

    Returns:
        tuple(numpy array, numpy array): texts and polarity

    """

    texts = []
    classes = []

    with open(filepath) as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            if line.startswith("#"):
                continue

            pieces = line.split("\t")
            if len(pieces) != 4:
                print("Warning: incorrect number of fields in the data file for line:", line)
                continue
            text = pieces[3]
            # remove start/end quotes
            text = text[1 : len(text) - 1]
            texts.append(text)

            the_class = []
            if pieces[2] == "n":
                the_class.append(1)
            else:
                the_class.append(0)
            if pieces[2] == "o":
                the_class.append(1)
            else:
                the_class.append(0)
            if pieces[2] == "p":
                the_class.append(1)
            else:
                the_class.append(0)
            classes.append(the_class)

    return np.asarray(texts), np.asarray(classes)


def load_dataseer_corpus_csv(filepath):
    """
    Load texts from the Dataseer dataset type corpus in csv format:

        doi,text,datatype,dataSubtype,leafDatatype

    Classification of the datatype follows a 3-level hierarchy, so the possible 3 classes are returned.
    dataSubtype and leafDatatype are optional

    Returns:
        tuple(numpy array, numpy array, numpy array, numpy array):
            texts, datatype, datasubtype, leaf datatype

    """
    df = pd.read_csv(filepath)
    df = df[pd.notnull(df["text"])]
    if "datatype" in df.columns:
        df = df[pd.notnull(df["datatype"])]
    if "reuse" in df.columns:
        df = df[pd.notnull(df["reuse"])]
    df.iloc[:, 1].fillna("NA", inplace=True)

    # shuffle, note that this is important for the reuse prediction, the following shuffle in place
    # and reset the index
    df = df.sample(frac=1).reset_index(drop=True)

    texts_list = []
    for j in range(0, df.shape[0]):
        texts_list.append(df.iloc[j, 1])

    if "reuse" in df.columns:
        # we simply get the reuse boolean value for the examples
        datareuses = df.iloc[:, 2]
        reuse_list = datareuses.values.tolist()
        reuse_list = np.asarray(reuse_list)

        # map boolean values to [0,1]
        def map_boolean(x):
            return [1.0, 0.0] if x == "no_reuse" else [0.0, 1.0]

        reuse_list = np.array(list(map(map_boolean, reuse_list)))
        print(reuse_list)
        return np.asarray(texts_list), reuse_list, None, None, ["no_reuse", "reuse"], None, None

    # otherwise we have the list of datatypes, and optionally subtypes and leaf datatypes
    datatypes = df.iloc[:, 2]
    datatypes_list = datatypes.values.tolist()
    datatypes_list = np.asarray(datatypes_list)
    datatypes_list_lower = np.char.lower(datatypes_list)
    list_classes_datatypes = np.unique(datatypes_list_lower)
    datatypes_final = normalize_classes(datatypes_list_lower, list_classes_datatypes)

    print(df.shape, df.shape[0], df.shape[1])

    if df.shape[1] > 3:
        # remove possible row with 'no_dataset'
        df = df[~df.datatype.str.contains("no_dataset")]
        datasubtypes = df.iloc[:, 3]
        datasubtypes_list = datasubtypes.values.tolist()
        datasubtypes_list = np.asarray(datasubtypes_list)
        datasubtypes_list_lower = np.char.lower(datasubtypes_list)
        list_classes_datasubtypes = np.unique(datasubtypes_list_lower)
        datasubtypes_final = normalize_classes(datasubtypes_list_lower, list_classes_datasubtypes)

    """
    if df.shape[1] > 4:
        leafdatatypes = df.iloc[:,4]
        leafdatatypes_list = leafdatatypes.values.tolist()
        leafdatatypes_list = np.asarray(leafdatatypes_list)
        #leafdatatypes_list_lower = np.char.lower(leafdatatypes_list)
        leafdatatypes_list_lower = leafdatatypes_list
        list_classes_leafdatatypes = np.unique(leafdatatypes_list_lower)
        print(list_classes_leafdatatypes)
        leafdatatypes_final = normalize_classes(leafdatatypes_list_lower, list_classes_leafdatatypes)
    """

    if df.shape[1] == 3:
        return np.asarray(texts_list), datatypes_final, None, None, list_classes_datatypes.tolist(), None, None
    # elif df.shape[1] == 4:
    else:
        return (
            np.asarray(texts_list),
            datatypes_final,
            datasubtypes_final,
            None,
            list_classes_datatypes.tolist(),
            list_classes_datasubtypes.tolist(),
            None,
        )
    """
    else:
        return np.asarray(texts_list), datatypes_final, datasubtypes_final, leafdatatypes_final, list_classes_datatypes.tolist(), list_classes_datasubtypes.tolist(), list_classes_leafdatatypes.tolist()
    """


def load_software_use_corpus_json(json_gz_file_path):
    """
    Load texts and classes from the corresponding Softcite corpus export in gzipped json format

    Classification of the software usage is binary

    Returns:
        tuple(numpy array, numpy array):
            texts, binary class (used/not_used)

    """

    texts_list = []
    classes_list = []

    with gzip.GzipFile(json_gz_file_path, "r") as fin:
        data = json.loads(fin.read().decode("utf-8"))
        if "documents" not in data:
            print("There is no usable classified text in the corpus file", json_gz_file_path)
            return None, None
        for document in data["documents"]:
            for segment in document["texts"]:
                if "entity_spans" in segment:
                    if "text" not in segment:
                        continue
                    text = segment["text"]
                    for entity_span in segment["entity_spans"]:
                        if entity_span["type"] == "software":
                            texts_list.append(text)
                            if "used" in entity_span and entity_span["used"]:
                                classes_list.append("used")
                            else:
                                classes_list.append("not_used")
    list_possible_classes = np.unique(classes_list)
    classes_list_final = normalize_classes(classes_list, list_possible_classes)

    texts_list_final = np.asarray(texts_list)

    texts_list_final, classes_list_final, _ = shuffle_triple_with_view(texts_list_final, classes_list_final)

    return texts_list_final, classes_list_final


def load_software_context_corpus_json(json_gz_file_path):
    """
    Load texts and classes from the corresponding Softcite mention corpus export in gzipped json format

    Classification of the software usage is multiclass/multilabel

    Returns:
        tuple(numpy array, numpy array):
            texts, classes_list

    """

    texts_list = []
    classes_list = []

    with gzip.GzipFile(json_gz_file_path, "r") as fin:
        data = json.loads(fin.read().decode("utf-8"))
        if "documents" not in data:
            print("There is no usable classified text in the corpus file", json_gz_file_path)
            return None, None
        for document in data["documents"]:
            for segment in document["texts"]:
                if "entity_spans" in segment:
                    if "text" not in segment:
                        continue
                    text = segment["text"]
                    for entity_span in segment["entity_spans"]:
                        if entity_span["type"] == "software":
                            texts_list.append(text)
                            classes = []
                            if "used" in entity_span and entity_span["used"]:
                                classes.append(1.0)
                            else:
                                classes.append(0.0)

                            if "created" in entity_span and entity_span["created"]:
                                classes.append(1.0)
                            else:
                                classes.append(0.0)

                            if "shared" in entity_span and entity_span["shared"]:
                                classes.append(1.0)
                            else:
                                classes.append(0.0)

                            classes_list.append(classes)

    # list_possible_classes = np.unique(classes_list)
    # classes_list_final = normalize_classes(classes_list, list_possible_classes)

    texts_list_final = np.asarray(texts_list)
    classes_list_final = np.asarray(classes_list)

    texts_list_final, classes_list_final, _ = shuffle_triple_with_view(texts_list_final, classes_list_final)

    return texts_list_final, classes_list_final


def load_software_dataset_context_corpus_json(json_gz_file_path):
    """
    Load texts and classes for software and dataset mention corpus KISH export in gzipped json format

    Classification of the dataset and software usage is multiclass/multilabel

    Returns:
        tuple(numpy array, numpy array):
            texts, classes_list

    """
    texts_list = []
    classes_list = []

    with gzip.GzipFile(json_gz_file_path, "r") as fin:
        data = json.loads(fin.read().decode("utf-8"))
        if "documents" not in data:
            print("There is no usable classified text in the corpus file", json_gz_file_path)
            return None, None
        for document in data["documents"]:
            for segment in document["texts"]:
                if "class_attributes" not in segment:
                    continue
                if "text" not in segment:
                    continue
                text = segment["text"]
                texts_list.append(text)
                classes = []
                classification = segment["class_attributes"]["classification"]
                if "used" in classification and classification["used"]["value"]:
                    classes.append(1.0)
                else:
                    classes.append(0.0)

                if "created" in classification and classification["created"]["value"]:
                    classes.append(1.0)
                else:
                    classes.append(0.0)

                if "shared" in classification and classification["shared"]["value"]:
                    classes.append(1.0)
                else:
                    classes.append(0.0)

                classes_list.append(classes)

    texts_list_final = np.asarray(texts_list)
    classes_list_final = np.asarray(classes_list)

    texts_list_final, classes_list_final, _ = shuffle_triple_with_view(texts_list_final, classes_list_final)

    return texts_list_final, classes_list_final


def normalize_classes(y, list_classes):
    """
    Replace string values of classes by their index in the list of classes
    """

    def f(x):
        return np.where(list_classes == x)

    intermediate = np.array([f(xi)[0] for xi in y])
    return np.array([vectorize(xi, len(list_classes)) for xi in intermediate])


def vectorize(index, size):
    """
    Create a numpy array of the provided size, where value at indicated index is 1, 0 otherwise
    """
    result = np.zeros(size)
    if index < size:
        result[index] = 1
    else:
        print("warning: index larger than vector size: ", index, size)
    return result



import logging
from functools import wraps

# derived from https://github.com/elifesciences/sciencebeam-trainer-delft/tree/develop/tests

LOGGER = logging.getLogger(__name__)


def log_on_exception(f: callable) -> callable:
    """
    Wraps function to log error on exception.
    That is useful for tests that log a lot of things,
    and pytest displaying the test failure at the top of the method.
    (there doesn't seem to be an option to change that)
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            LOGGER.exception("failed due to %s", repr(e))
            raise

    return wrapper



import logging

from delft.utilities.Tokenizer import tokenizeAndFilter


class TestTokenizer:
    LOGGER = logging.getLogger(__name__)

    def test_tokenize_and_filter(self):
        tokens, offsets = tokenizeAndFilter("this is a simple text")

        assert len(tokens) == len(offsets)
        assert len(tokens) == 5



from delft.utilities.Tokenizer import tokenizeAndFilter, tokenizeAndFilterSimple


class TestTokenizer:
    def test_tokenizer_filter_simple(self):
        input = "this is a test, but a stupid test!!"

        output = tokenizeAndFilterSimple(input)

        assert len(output) == 11
        assert output == ["this", "is", "a", "test", ",", "but", "a", "stupid", "test", "!", "!"]

    def test_tokenizer_filter(self):
        input = "this is a test, but a stupid test!!"

        output = tokenizeAndFilter(input)

        assert len(output) == 2
        assert output[0] == ["this", "is", "a", "test", ",", "but", "a", "stupid", "test", "!", "!"]
        assert output[1] == [
            (0, 4),
            (5, 7),
            (8, 9),
            (10, 14),
            (14, 15),
            (16, 19),
            (20, 21),
            (22, 28),
            (29, 33),
            (33, 34),
            (34, 35),
        ]

    def test_tokenizer_filter_simple_with_breaklines(self):
        input = "\nthis is yet \u2666 another, dummy... test,\na [stupid] test?!"

        output = tokenizeAndFilterSimple(input)

        assert len(output) == 19
        assert output == [
            "this",
            "is",
            "yet",
            "\u2666",
            "another",
            ",",
            "dummy",
            ".",
            ".",
            ".",
            "test",
            ",",
            "a",
            "[",
            "stupid",
            "]",
            "test",
            "?",
            "!",
        ]

    def test_tokenizer_filter_with_breaklines(self):
        input = "\nthis is yet \u2666 another, dummy... test,\na [stupid] test?!"

        output = tokenizeAndFilter(input)

        assert len(output) == 2
        assert output[0] == [
            "this",
            "is",
            "yet",
            "\u2666",
            "another",
            ",",
            "dummy",
            ".",
            ".",
            ".",
            "test",
            ",",
            "a",
            "[",
            "stupid",
            "]",
            "test",
            "?",
            "!",
        ]
        assert output[1] == [
            (1, 5),
            (6, 8),
            (9, 12),
            (13, 14),
            (15, 22),
            (22, 23),
            (24, 29),
            (29, 30),
            (30, 31),
            (31, 32),
            (33, 37),
            (37, 38),
            (39, 40),
            (41, 42),
            (42, 48),
            (48, 49),
            (50, 54),
            (54, 55),
            (55, 56),
        ]
